import os
import math

import numpy as np
import pandas as pd

import celescope.tools
from celescope.capture_virus.otsu import array2hist, makePlot, threshold_otsu
from celescope.tools.analysis_mixin import AnalysisMixin
from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils

TOOLS_DIR = os.path.dirname(celescope.tools.__file__)


@utils.add_log
def analysis_capture_virus(args):

    step_name = 'analysis_capture_virus'
    runner = Analysis_capture_virus(args, step_name)
    runner.run()


def get_opts_analysis_capture_virus(parser, sub_program):
    parser.add_argument("--umi_threshold", help='method to find virus UMI threshold',
                        choices=['otsu', 'auto', 'none'], default='otsu')
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='match_dir', required=True)
        parser.add_argument('--virus_file', help='virus UMI count file', required=True)


class Analysis_capture_virus(Step, AnalysisMixin):

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        AnalysisMixin.__init__(self, args)
        self.virus_file = args.virus_file
        self.umi_threshold = args.umi_threshold

        # parse
        _barcodes, self.n_cell = utils.read_barcode_file(self.match_dir)

        # set
        self.threshold = "None"
        self.add_metric(
            name='Number of cells',
            value=self.n_cell,
        )

        # out
        self.virus_tsne_file = f'{self.outdir}/{self.sample}_virus_tsne.tsv'
        self.auto_virus_file = f'{self.outdir}/{self.sample}_auto_UMI_count.tsv'
        self.otsu_virus_file = f'{self.outdir}/{self.sample}_otsu_UMI_count.tsv'
        self.otsu_plot = f'{self.outdir}/{self.sample}_otsu_plot.png'
    
    def run(self):
        if self.umi_threshold == 'otsu':
            self.otsu_threshold()
            virus_file = self.otsu_virus_file
        elif self.umi_threshold == 'auto':
            self.auto_threshold()
            virus_file = self.auto_virus_file
        else:
            virus_file = self.virus_file
        virus_df = pd.read_csv(virus_file, sep="\t")

        cluster_tsne = self.get_cluster_tsne(colname='cluster', tsne_df=self.tsne_df)
        virus_tsne = self.get_virus_tsne(virus_df)
        table_dict = self.get_table_dict(self.get_marker_table())

        self.add_data_item(cluster_tsne=cluster_tsne)
        self.add_data_item(virus_tsne=virus_tsne)
        self.add_data_item(table_dict=table_dict)
        self.clean_up()

    def get_virus_tsne(self, virus_df):
        virus_tsne_df = pd.merge(self.tsne_df, virus_df, on="barcode", how="left")
        virus_tsne_df.to_csv(self.virus_tsne_file, sep='\t')
        
        #get the sum of virus
        total_virus = virus_tsne_df.loc[:,"tag"].count()
        #get the sum of virus in each cluster
        cluster_virus = virus_tsne_df.groupby("cluster")["tag"].count()
        #Total cell each cluster
        total_cell_cluster = virus_tsne_df.groupby("cluster")["barcode"].count()
        total_cell = sum(total_cell_cluster)
        #add sum of virus to metric
        self.add_metric(name = 'Number of Cells with Virus',
                        value = total_virus,
                        total = total_cell)
        cluster = 1    
        for num_virus in cluster_virus:
            cluster_cell_num = total_cell_cluster[cluster]
            self.add_metric(name = f'Cluster_{cluster}',
                            value = num_virus,
                            total = cluster_cell_num)
            cluster += 1


        virus_tsne_df["UMI"] = virus_tsne_df["UMI"].fillna(0)
        tSNE_1 = list(virus_tsne_df.tSNE_1)
        tSNE_2 = list(virus_tsne_df.tSNE_2)
        virus_UMI = list(virus_tsne_df.UMI)
        res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "virus_UMI": virus_UMI}
        return res

    @utils.add_log
    def otsu_threshold(self):
        df_virus = pd.read_csv(self.virus_file, sep='\t')
        array = np.log10(df_virus["UMI"])
        hist = array2hist(array)
        thresh = threshold_otsu(hist)
        makePlot(hist, thresh, self.otsu_plot)

        threshold = int(10 ** thresh)
        self.add_metric(
            name='Otsu UMI Threshold',
            value=threshold,
        )
        df_thresh = df_virus[df_virus["UMI"] >= threshold]
        df_thresh.to_csv(self.otsu_virus_file, sep='\t')

    @utils.add_log
    def auto_threshold(self):
        """
        threhold = 99 percentile of all cell virus UMIs / 10
        """
        df_virus = pd.read_csv(self.virus_file, sep='\t')
        umis = df_virus["UMI"]

        cell_99th = self.n_cell // 100
        if cell_99th >= len(umis):
            threshold = 1
        else:
            sorted_umis = sorted(umis, reverse=True)
            percentile_99_umi = sorted_umis[cell_99th]
            threshold = math.ceil(percentile_99_umi / 10)
        self.add_metric(
            name='Auto UMI Threshold',
            value=threshold,
        )
        df_threshold = df_virus[df_virus["UMI"] >= threshold]
        df_threshold.to_csv(self.auto_virus_file, sep='\t')



        




import matplotlib.pyplot as plt
import matplotlib
import numpy as np

matplotlib.use('Agg')


def threshold_otsu(hist):
    """Return threshold value based on Otsu's method.

    hist : array, or 2-tuple of arrays, optional
        Histogram from which to determine the threshold, and optionally a
        corresponding array of bin center intensities.
        An alternative use of this function is to pass it only hist.
    Returns
    -------
    threshold : float
        Upper threshold value. All pixels with an intensity higher than
        this value are assumed to be foreground.
    References
    ----------
    .. [1] Wikipedia, https://en.wikipedia.org/wiki/Otsu's_Method
    """
    counts, bin_centers = hist
    bin_centers = bin_centers[:-1]

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]
    # class means for all possible thresholds
    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of ``weight1``/``mean1`` should pair with zero values in
    # ``weight2``/``mean2``, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2
    if len(variance12) == 0:
        return 0
    idx = np.nanargmax(variance12)
    threshold = bin_centers[idx]

    return threshold


def array2hist(array, binWidth=0.2):
    counts, bins = np.histogram(array, bins=np.arange(0, max(array)+binWidth, binWidth))
    return counts, bins


def makePlot(hist, thresh, fname):
    counts, bins = hist
    plt.hist(bins[:-1], bins, weights=counts)
    plt.axvline(thresh, color='r')
    plt.savefig(fname)
    plt.close()



"""
cellranger3 cell-calling on raw matrix
"""

import os
import sys

import numpy as np
import pandas as pd
import scipy.io

from celescope.tools.__init__ import (BARCODE_FILE_NAME, FEATURE_FILE_NAME,
                                      MATRIX_FILE_NAME)
from celescope.tools.cellranger3.cell_calling_3 import find_nonambient_barcodes
import celescope.tools.utils as utils


EXPECTED_CELL_NUM = 3000


@utils.add_log
def read_raw_matrix(all_matrix_10X_dir):

    raw_mat_path = os.path.join(all_matrix_10X_dir, MATRIX_FILE_NAME)
    raw_mat = scipy.io.mmread(raw_mat_path)  # scipy.sparse.coo.coo_matrix

    raw_features_path = os.path.join(all_matrix_10X_dir, FEATURE_FILE_NAME)

    raw_barcodes_path = os.path.join(all_matrix_10X_dir, BARCODE_FILE_NAME)
    raw_barcodes_df = pd.read_csv(raw_barcodes_path, sep='\t', error_bad_lines=False, names=['barcode'])
    raw_barcodes = np.array(raw_barcodes_df['barcode'].tolist())

    return raw_mat, raw_features_path, raw_barcodes


@utils.add_log
def cell_calling(raw_mat, expected_cell_num):

    filtered_bc_indices, _round_1_filtered_metrics, _non_ambient_barcode_result = find_nonambient_barcodes(
        raw_mat=raw_mat, recovered_cells=expected_cell_num)

    return filtered_bc_indices

class Cell_calling():
    def __init__(self, outdir, raw_mat, raw_features_path, raw_barcodes):
        self.raw_mat = raw_mat
        self.raw_features_path = raw_features_path
        self.raw_barcodes = raw_barcodes
        self.expected_cell_num = EXPECTED_CELL_NUM
        self.out_mat = f'{outdir}/{MATRIX_FILE_NAME}'
        self.out_feature = f'{outdir}/{FEATURE_FILE_NAME}'
        self.out_barcode = f'{outdir}/{BARCODE_FILE_NAME}'
        utils.check_mkdir(outdir)
        
    def write_slice_matrix(self, filtered_bc_indices):
        mtx_csc = self.raw_mat.tocsc()
        sliced_mtx = mtx_csc[:,filtered_bc_indices]
        scipy.io.mmwrite(target=self.out_mat, a=sliced_mtx)

        barcodes = pd.Series(self.raw_barcodes[filtered_bc_indices])
        barcodes.to_csv(self.out_barcode, index=False, sep='\t', header=False)

        os.system(f"cp {self.raw_features_path} {self.out_feature}")

    @utils.add_log
    def run(self):
        filtered_bc_indices = cell_calling(self.raw_mat, self.expected_cell_num)
        self.write_slice_matrix(filtered_bc_indices)


def main():
    outdir = sys.argv[1]
    all_matrix_10X_dir = sys.argv[2]
    raw_mat, raw_features_path, raw_barcodes = read_raw_matrix(all_matrix_10X_dir)
    runner = Cell_calling(outdir, raw_mat, raw_features_path, raw_barcodes)
    runner.run()

if __name__ == '__main__':
    main()


from celescope.tools.multi import Multi
from celescope.vdj.__init__ import __ASSAY__


class Multi_vdj(Multi):
    """
    Usage
    ```
    multi_vdj \\
        --mapfile ./vdj.mapfile \\
        --type TCR \\
        --thread 8 \\
        --mod shell
    ``` 
    """

    def mapping_vdj(self, sample):
        step = 'mapping_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        if self.args.not_consensus:
            fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        else:
            fq = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fq'
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=self.args.thread)

    def count_vdj(self, sample):
        # count_vdj
        step = 'count_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        UMI_count_filter_file = (
            f'{self.outdir_dic[sample]["mapping_vdj"]}/{sample}_UMI_count_filtered.tsv'
        )
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--UMI_count_filter_file {UMI_count_filter_file} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=self.args.thread)


def main():
    multi = Multi_vdj(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import argparse
import glob
import gzip
import importlib
import itertools
import json
import logging
import os
import re
import resource
import subprocess
import time
from collections import Counter, defaultdict
from datetime import timedelta
from functools import wraps

import numpy as np
import pandas as pd
import pysam
import xopen

import celescope.tools
from celescope.tools.__init__ import PATTERN_DICT
from celescope.__init__ import ROOT_PATH
from celescope.capture_virus.otsu import array2hist, makePlot, threshold_otsu

tools_dir = os.path.dirname(celescope.tools.__file__)


def add_log(func):
    '''
    logging start and done.
    '''
    logFormatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    module = func.__module__
    name = func.__name__
    logger_name = f'{module}.{name}'
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    logger.addHandler(consoleHandler)

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args and hasattr(args[0], 'debug') and args[0].debug:
            logger.setLevel(10)  # debug

        logger.info('start...')
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        used = timedelta(seconds=end - start)
        logger.info('done. time used: %s', used)
        return result

    wrapper.logger = logger
    return wrapper


def using(point=""):
    usage = resource.getrusage(resource.RUSAGE_SELF)
    return '''%s: usertime=%s systime=%s mem=%s mb
        ''' % (point, usage[0], usage[1],
               usage[2]/1024.0)


def add_mem(func):
    '''
    logging mem.
    '''
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    module = func.__module__
    name = func.__name__
    logger_name = f'{module}.{name}'
    logger = logging.getLogger(logger_name)

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(using("before"))
        result = func(*args, **kwargs)
        logger.info(using("after"))
        return result

    wrapper.logger = logger
    return wrapper


def arg_str(arg, arg_name):
    '''
    return action store_true arguments as string
    '''
    if arg:
        return '--' + arg_name
    return ''


def format_stat(count, total_count):
    percent = round(count / total_count * 100, 2)
    string = f'{format_number(count)}({percent}%)'
    return string


def multi_opts(assay):
    readme = f'{assay} multi-samples'
    parser = argparse.ArgumentParser(readme)
    parser.add_argument('--mod', help='mod, sjm or shell', choices=['sjm', 'shell'], default='sjm')
    parser.add_argument(
        '--mapfile',
        help='''
            tsv file, 4 columns:
            1st col: LibName;
            2nd col: DataDir;
            3rd col: SampleName;
            4th col: Cell number or match_dir, optional;
        ''',
        required=True)
    parser.add_argument('--chemistry', choices=list(PATTERN_DICT.keys()), help='chemistry version', default='auto')
    parser.add_argument('--whitelist', help='cellbarcode list')
    parser.add_argument('--linker', help='linker')
    parser.add_argument('--pattern', help='read1 pattern')
    parser.add_argument('--outdir', help='output dir', default="./")
    parser.add_argument(
        '--adapt',
        action='append',
        help='adapter sequence',
        default=[
            'polyT=A{15}',
            'p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC'])
    parser.add_argument(
        '--minimum_length',
        dest='minimum_length',
        help='minimum_length',
        default=20)
    parser.add_argument(
        '--nextseq-trim',
        dest='nextseq_trim',
        help='nextseq_trim',
        default=20)
    parser.add_argument(
        '--overlap',
        help='minimum overlap length, default=5',
        default=5)
    parser.add_argument(
        '--lowQual',
        type=int,
        help='max phred of base as lowQual',
        default=0)
    parser.add_argument(
        '--lowNum',
        type=int,
        help='max number with lowQual allowed',
        default=2)
    parser.add_argument(
        '--rm_files',
        action='store_true',
        help='remove redundant fq.gz and bam after running')
    parser.add_argument('--steps_run', help='steps to run', default='all')
    parser.add_argument('--debug', help='debug or not', action='store_true')
    parser.add_argument('--outFilterMatchNmin', help='STAR outFilterMatchNmin', default=0)
    return parser


def link_data(outdir, fq_dict):
    raw_dir = f'{outdir}/data_give/rawdata'
    os.system('mkdir -p %s' % (raw_dir))
    with open(raw_dir + '/ln.sh', 'w') as fh:
        fh.write('cd %s\n' % (raw_dir))
        for s, arr in fq_dict.items():
            fh.write('ln -sf %s %s\n' % (arr[0], s + '_1.fq.gz'))
            fh.write('ln -sf %s %s\n' % (arr[1], s + '_2.fq.gz'))


def generic_open(file_name, *args, **kwargs):
    if file_name.endswith('.gz'):
        file_obj = gzip.open(file_name, *args, **kwargs)
    else:
        file_obj = open(file_name, *args, **kwargs)
    return file_obj


@add_log
def get_id_name_dict(gtf_file):
    """
    get gene_id:gene_name from gtf file
        - one gene_name with multiple gene_id: "_{count}" will be added to gene_name.
        - one gene_id with multiple gene_name: error.
        - duplicated (gene_name, gene_id): ignore duplicated records and print a warning.

    Returns:
        {gene_id: gene_name} dict
    """

    gene_id_pattern = re.compile(r'gene_id "(\S+)";')
    gene_name_pattern = re.compile(r'gene_name "(\S+)"')
    id_name = {}
    c = Counter()
    with generic_open(gtf_file) as f:
        for line in f:
            if not line.strip():
                continue
            if line.startswith('#'):
                continue
            tabs = line.split('\t')
            gtf_type, attributes = tabs[2], tabs[-1]
            if gtf_type == 'gene':
                gene_id = gene_id_pattern.findall(attributes)[-1]
                gene_names = gene_name_pattern.findall(attributes)
                if not gene_names:
                    gene_name = gene_id
                else:
                    gene_name = gene_names[-1]
                c[gene_name] += 1
                if c[gene_name] > 1:
                    if gene_id in id_name:
                        assert id_name[gene_id] == gene_name, (
                            'one gene_id with multiple gene_name '
                            f'gene_id: {gene_id}, '
                            f'gene_name this line: {gene_name}'
                            f'gene_name previous line: {id_name[gene_id]}'
                        )
                        get_id_name_dict.logger.warning(
                            'duplicated (gene_id, gene_name)'
                            f'gene_id: {gene_id}, '
                            f'gene_name {gene_name}'
                        )
                        c[gene_name] -= 1
                    else:
                        gene_name = f'{gene_name}_{c[gene_name]}'
                id_name[gene_id] = gene_name
    return id_name


@add_log
def process_read(
        read2_file, pattern_dict, barcode_dict, linker_dict,
        barcode_length, linker_length):

    # if valid, return (True)
    metrics = defaultdict(int)
    res_dict = genDict(dim=3)
    read2 = xopen.xopen(read2_file, "rt")
    while True:
        line1 = read2.readline()
        line2 = read2.readline()
        read2.readline()
        line4 = read2.readline()
        if not line4:
            break
        metrics['Total Reads'] += 1
        attr = str(line1).strip("@").split("_")
        barcode = str(attr[0])
        umi = str(attr[1])
        seq = line2.strip()
        if linker_length != 0:
            seq_linker = seq_ranges(seq, pattern_dict['L'])
            if len(seq_linker) < linker_length:
                metrics['Reads Unmapped too Short'] += 1
                continue
        if barcode_dict:
            seq_barcode = seq_ranges(seq, pattern_dict['C'])
            if barcode_length != len(seq_barcode):
                miss_length = barcode_length - len(seq_barcode)
                if miss_length > 2:
                    metrics['Reads Unmapped too Short'] += 1
                    continue
                seq_barcode = seq_barcode + "A" * miss_length

        # check linker
        if linker_length != 0:
            valid_linker = False
            for linker_name in linker_dict:
                if hamming_correct(linker_dict[linker_name], seq_linker):
                    valid_linker = True
                    break
        else:
            valid_linker = True

        if not valid_linker:
            metrics['Reads Unmapped Invalid Linker'] += 1
            continue

        # check barcode
        valid_barcode = False
        for barcode_name in barcode_dict:
            if hamming_correct(barcode_dict[barcode_name], seq_barcode):
                res_dict[barcode][barcode_name][umi] += 1
                valid_barcode = True
                break

        if not valid_barcode:
            metrics['Reads Unmapped Invalid Barcode'] += 1
            continue

        # mapped
        metrics['Reads Mapped'] += 1
        if metrics['Reads Mapped'] % 1000000 == 0:
            process_read.logger.info(str(metrics['Reads Mapped']) + " reads done.")

    return res_dict, metrics


def seq_ranges_exception(seq, pattern_dict):
    # get subseq with intervals in arr and concatenate
    length = len(seq)
    for x in pattern_dict:
        if length < x[1]:
            raise Exception(f'invalid seq range {x[0]}:{x[1]} in read')
    return ''.join([seq[x[0]:x[1]]for x in pattern_dict])


def seq_ranges(seq, pattern_dict):
    # get subseq with intervals in arr and concatenate
    return ''.join([seq[x[0]:x[1]]for x in pattern_dict])


def read_one_col(file):
    df = pd.read_csv(file, header=None)
    col1 = list(df.iloc[:, 0])
    num = len(col1)
    return col1, num

def get_gene_region_from_bed(panel):
    """
    Returns 
    - genes
    - position_df with 'Chromosome', 'Start', 'End'
    """
    file_path = f'{ROOT_PATH}/data/snp/panel/{panel}.bed'
    bed_file_df = pd.read_table(file_path, 
                                usecols=[0,1,2,3],
                                names=['Chromosome', 'Start', 'End','Gene'],
                                sep = "\t")
    position_df = bed_file_df.loc[:,['Chromosome', 'Start', 'End']]
    genes = set(bed_file_df.loc[:,'Gene'].to_list())
    return genes, position_df

def read_fasta(fasta_file, equal=False):
    # seq must have equal length
    fa_dict = {}
    length = None
    with pysam.FastxFile(fasta_file) as infile:
        for index, record in enumerate(infile):
            seq = record.sequence
            if index == 0:
                length = len(seq)
            if equal:
                if length != len(seq):
                    raise Exception(f"{fasta_file} have different seq length")
            fa_dict[record.name] = seq
    return fa_dict, length


def hamming_correct(string1, string2):
    threshold = len(string1) / 10 + 1
    if hamming_distance(string1, string2) < threshold:
        return True
    return False


def hamming_distance(string1, string2):
    distance = 0
    length = len(string1)
    length2 = len(string2)
    if (length != length2):
        raise Exception(f"string1({length}) and string2({length2}) do not have same length")
    for i in range(length):
        if string1[i] != string2[i]:
            distance += 1
    return distance


def gen_stat(df, stat_file):
    # 3cols: item count total_count

    def add_percent(row):
        count = row['count']
        percent = count / row['total_count']
        value = f'{format_number(count)}({round(percent * 100, 2)}%)'
        return value

    df.loc[:, 'value'] = df.loc[:, 'count']
    df.loc[~df['total_count'].isna(), 'value'] = df.loc[~df['total_count'].isna(), :].apply(
        add_percent, axis=1
    )
    df.loc[df['total_count'].isna(), 'value'] = df.loc[df['total_count'].isna(), :].apply(
        lambda row: f'{format_number(row["count"])}', axis=1
    )
    df = df.loc[:, ["item", "value"]]
    df.to_csv(stat_file, sep=":", header=None, index=False)


def get_read(library_id, library_path, read='1'):
    read1_list = [f'_{read}', f'R{read}', f'R{read}_001']
    fq_list = ['fq', 'fastq']
    suffix_list = ["", ".gz"]
    read_pattern_list = [
        f'{library_path}/*{library_id}*{read}.{fq_str}{suffix}'
        for read in read1_list
        for fq_str in fq_list
        for suffix in suffix_list
    ]
    fq_list = [glob.glob(read1_pattern) for read1_pattern in read_pattern_list]
    fq_list = sorted(non_empty for non_empty in fq_list if non_empty)
    fq_list = list(itertools.chain(*fq_list))
    if len(fq_list) == 0:
        print("Allowed R1 patterns:")
        for pattern in read_pattern_list:
            print(pattern)
        raise Exception(
            '\n'
            f'Invalid Read{read} path! \n'
            f'library_id: {library_id}\n'
            f'library_path: {library_path}\n'
        )
    return fq_list


def get_fq(library_id, library_path):
    fq1_list = get_read(library_id, library_path, read='1')
    fq2_list = get_read(library_id, library_path, read='2')
    if len(fq1_list) != len(fq2_list):
        raise Exception("Read1 and Read2 fastq number do not match!")
    fq1 = ",".join(fq1_list)
    fq2 = ",".join(fq2_list)
    return fq1, fq2


def format_number(number: int) -> str:
    return format(number, ",")


def format_metrics(metrics: dict):
    for key in metrics:
        value = metrics[key]
        metrics[key] = format_number(value)
        if isinstance(value, float):
            metrics[key] = round(value, 2)


def format_ratios(ratios: dict):
    for key in ratios:
        ratios[key] = round(ratios[key] * 100, 2)


def get_slope(x, y, window=200, step=10):
    assert len(x) == len(y)
    start = 0
    last = len(x)
    res = [[], []]
    while True:
        end = start + window
        if end > last:
            break
        z = np.polyfit(x[start:end], y[start:end], 1)
        res[0].append(x[start])
        res[1].append(z[0])
        start += step
    return res


def genDict(dim=3, valType=int):
    if dim == 1:
        return defaultdict(valType)
    else:
        return defaultdict(lambda: genDict(dim - 1, valType=valType))


def report_prepare(outdir, **kwargs):
    json_file = outdir + '/../.data.json'
    if not os.path.exists(json_file):
        data = {}
    else:
        fh = open(json_file)
        data = json.load(fh)
        fh.close()

    for key in kwargs:
        data[key] = kwargs[key]

    with open(json_file, 'w') as fh:
        json.dump(data, fh)


def parse_vcf(vcf_file, cols=('chrom', 'pos', 'alleles'), infos=('VID', 'CID')):
    vcf = pysam.VariantFile(vcf_file)
    df = pd.DataFrame(columns=[col.capitalize() for col in cols] + infos)
    rec_dict = {}
    for rec in vcf.fetch():

        for col in cols:
            rec_dict[col.capitalize()] = getattr(rec, col)
            if col == 'alleles':
                rec_dict['Alleles'] = '-'.join(rec_dict['Alleles'])

        for info in infos:
            rec_dict[info] = rec.info[info]

        '''
        rec_dict['GT'] = [s['GT'] for s in rec.samples.values()][0]
        rec_dict['GT'] = [str(item) for item in rec_dict['GT']]
        rec_dict['GT'] = '/'.join(rec_dict['GT'])
        '''

        df = df.append(pd.Series(rec_dict), ignore_index=True)
    vcf.close()
    return df


def parse_annovar(annovar_file):
    df = pd.DataFrame(columns=['Gene', 'mRNA', 'Protein', 'COSMIC'])
    with open(annovar_file, 'rt') as f:
        index = 0
        for line in f:
            index += 1
            if index == 1:
                continue
            attrs = line.split('\t')
            gene = attrs[6]
            func = attrs[5]
            if func == 'exonic':
                changes = attrs[9]
                cosmic = attrs[10]
            else:
                changes = attrs[7]
                cosmic = attrs[8]
            change_list = list()
            for change in changes.split(','):
                change_attrs = change.split(':')
                mRNA = ''
                protein = ''
                for change_index in range(len(change_attrs)):
                    change_attr = change_attrs[change_index]
                    if change_attr.startswith('c.'):
                        base = change_attr.strip('c.')
                        exon = change_attrs[change_index - 1]
                        mRNA = f'{exon}:{base}'
                    if change_attr.startswith('p.'):
                        protein = change_attr.strip('p.')
                if not (mRNA, protein) in change_list:
                    change_list.append((mRNA, protein))
            combine = [','.join(item) for item in list(zip(*change_list))]
            mRNA = combine[0]
            protein = combine[1]
            df = df.append({
                'Gene': gene,
                'mRNA': mRNA,
                'Protein': protein,
                'COSMIC': cosmic,
            }, ignore_index=True)
    return df


def read_barcode_file(match_dir, return_file=False):
    '''
    multi version compatible
    '''
    match_barcode_file1 = glob.glob(
        f"{match_dir}/*count*/*_cellbarcode.tsv")
    match_barcode_file2 = glob.glob(
        f"{match_dir}/*count*/*matrix_10X/*_cellbarcode.tsv")
    match_barcode_file3 = glob.glob(
        f"{match_dir}/*count*/*matrix_10X/*barcodes.tsv")
    match_barcode_file = (
        match_barcode_file1 +
        match_barcode_file2 +
        match_barcode_file3)[0]
    match_barcode, cell_total = read_one_col(match_barcode_file)
    if return_file:
        return match_barcode, (cell_total, match_barcode_file)
    return match_barcode, cell_total


def get_barcodes_from_matrix_dir(matrix_dir):
    barcodes_file = f'{matrix_dir}/barcodes.tsv'
    match_barcode, _cell_total = read_one_col(barcodes_file)
    return match_barcode


def parse_match_dir(match_dir):
    match_dict = {}
    match_barcode, cell_total = read_barcode_file(match_dir)
    match_dict['match_barcode'] = match_barcode
    match_dict['cell_total'] = cell_total
    match_dict['matrix_dir'] = glob.glob(f'{match_dir}/*count*/*matrix_10X')[0]
    match_dict['tsne_coord'] = glob.glob(f'{match_dir}/*analysis*/*tsne_coord.tsv')[0]
    match_dict['markers'] = glob.glob(f'{match_dir}/*analysis*/*markers.tsv')[0]
    try:
        match_dict['rds'] = glob.glob(f'{match_dir}/*analysis/*.rds')[0]
    except IndexError:
        match_dict['rds'] = None
    return match_dict


@add_log
def STAR_util(
    sample,
    outdir,
    input_read,
    genomeDir,
    runThreadN,
    outFilterMatchNmin=35,
    out_unmapped=False,
    outFilterMultimapNmax=1,
    outBAMsortingBinsN=50,
):

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    out_prefix = f'{outdir}/{sample}_'
    out_BAM = out_prefix + "Aligned.sortedByCoord.out.bam"

    cmd = f"STAR \
 --genomeDir {genomeDir} \
 --readFilesIn {input_read}\
 --readFilesCommand zcat\
 --outSAMtype BAM SortedByCoordinate\
 --runThreadN {runThreadN}\
 --outFilterMatchNmin {outFilterMatchNmin}\
 --outFileNamePrefix {out_prefix}\
 --outFilterMultimapNmax {outFilterMultimapNmax}\
 --outBAMsortingBinsN {outBAMsortingBinsN}\
 --limitBAMsortRAM 100000000000 "

    if out_unmapped:
        cmd += ' --outReadsUnmapped Fastx '

    STAR_util.logger.info(cmd)
    subprocess.check_call(cmd, shell=True)

    cmd = "samtools index {out_BAM}".format(out_BAM=out_BAM)
    STAR_util.logger.info(cmd)
    os.system(cmd)


def get_scope_bc(bctype):
    root_path = os.path.dirname(celescope.__file__)
    linker_f = glob.glob(f'{root_path}/data/chemistry/{bctype}/linker*')[0]
    whitelist_f = f'{root_path}/data/chemistry/{bctype}/bclist'
    return linker_f, whitelist_f


def fastq_line(name, seq, qual):
    return f'@{name}\n{seq}\n+\n{qual}\n'


def s_common(parser):
    """subparser common arguments
    """
    parser.add_argument('--outdir', help='output dir', required=True)
    parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument('--sample', help='sample name', required=True)
    parser.add_argument('--thread', default=4)
    parser.add_argument('--debug', help='debug', action='store_true')
    return parser


def find_assay_init(assay):
    init_module = importlib.import_module(f"celescope.{assay}.__init__")
    return init_module


def find_step_module(assay, step):
    init_module = find_assay_init(assay)
    try:
        step_module = importlib.import_module(f"celescope.{assay}.{step}")
    except ModuleNotFoundError:
        try:
            step_module = importlib.import_module(f"celescope.tools.{step}")
        except ModuleNotFoundError:
            module_path = init_module.IMPORT_DICT[step]
            step_module = importlib.import_module(f"{module_path}.{step}")

    return step_module


def find_step_module_with_folder(assay, step):
    init_module = find_assay_init(assay)
    folder = ""
    try:
        step_module = importlib.import_module(f"celescope.{assay}.{step}")
        folder = assay
    except ModuleNotFoundError:
        try:
            step_module = importlib.import_module(f"celescope.tools.{step}")
            folder = 'tools'
        except ModuleNotFoundError:
            module_path = init_module.IMPORT_DICT[step]
            step_module = importlib.import_module(f"{module_path}.{step}")
            folder = module_path.split('.')[1]

    return step_module, folder


def sort_bam(input_bam, output_bam, threads=1):
    cmd = (
        f'samtools sort {input_bam} '
        f'-o {output_bam} '
        f'--threads {threads} '
    )
    subprocess.check_call(cmd, shell=True)


def index_bam(input_bam):
    cmd = f"samtools index {input_bam}"
    subprocess.check_call(cmd, shell=True)


def check_mkdir(dir_name):
    if not os.path.exists(dir_name):
        os.system(f"mkdir -p {dir_name}")


def otsu_min_support_read(array, otsu_plot):
    """
    get otsu threshold and plot
    """
    array = np.log10(array)
    hist = array2hist(array)
    thresh = threshold_otsu(hist)
    makePlot(hist, thresh, otsu_plot)
    threshold = round(10 ** thresh,1)
    return threshold


class Samtools():
    def __init__(self, in_bam, out_bam, threads=1):
        self.in_bam = in_bam
        self.out_bam = out_bam
        self.threads = threads

    def sort_bam(self, by='coord', print_log=False):
        cmd = f"samtools sort {self.in_bam} -o {self.out_bam} --threads {self.threads}"
        if by == "name":
            cmd += " -n"
        if print_log:
            print(cmd)
        subprocess.check_call(cmd, shell=True)
        return cmd




import argparse
import inspect
import os

import celescope.tools.utils as utils
from celescope.celescope import ArgFormatter
from celescope.__init__ import ASSAY_DICT, RELEASED_ASSAYS

PRE_PROCESSING_STEPS = ('sample', 'barcode', 'cutadapt')
DOCS_DIR = f'docs/'
TEMPLATE_DIR = 'docs_template/'
MANUAL = f'{DOCS_DIR}/manual.md'


def get_argument_docs_from_parser(parser):
    argument_docs = ""
    for argument in parser._option_string_actions:
        if not argument in ['-h', '--help']:
            help_msg = parser._option_string_actions[argument].help
            if help_msg:
                help_msg = help_msg.strip()
            argument_docs += (f'`{argument}` {help_msg}\n\n')
    argument_docs = "## Arguments\n" + argument_docs
    return argument_docs


def get_class_docs(step_module):
    titles = ("Features", "Output", "Usage")
    class_docs = ""
    for child in inspect.getmembers(step_module, inspect.isclass):
        """Filter out class not defined in step_module"""
        class_obj = child[1]
        if class_obj.__module__ != step_module.__name__:
            continue
        doc = inspect.getdoc(class_obj)
        
        # only write class docs when the doc contains at least one title
        write_bool = False
        if doc:
            for title in titles:
                if title in doc:
                    write_bool = True

        if write_bool:
            for line in doc.split('\n'):
                for title in titles:
                    if line.find(title) != -1:
                        class_docs += "## " + line + '\n'
                        break
                else:
                    class_docs += line + '\n'
    class_docs += '\n\n'
    return class_docs


class Docs():
    def __init__(self, assay):
        self.assay = assay

        init_module = utils.find_assay_init(assay)
        self.steps = init_module.__STEPS__.copy()
        self.steps.append(f'multi_{assay}')

        self.out_md_dict = {}
        self.relative_md_path = {}
        self.release_bool = self.assay in RELEASED_ASSAYS

        assay_dir = f'docs/{assay}'
        if not os.path.exists(assay_dir):
            os.system(f'mkdir -p {assay_dir}') 
        tools_dir = 'docs/tools'
        if not os.path.exists(tools_dir):
            os.system(f'mkdir {tools_dir}')

    @utils.add_log
    def get_argument_docs(self, step, step_module):
        self.get_argument_docs.logger.info(step)
        if step.startswith("multi"):
            multi_class = getattr(step_module, f'Multi_{self.assay}')
            multi_obj = multi_class(self.assay)
            argument_docs = get_argument_docs_from_parser(multi_obj.parser)
        else:
            parser = argparse.ArgumentParser(description='CeleScope', formatter_class=ArgFormatter)
            func_opts = getattr(step_module, f"get_opts_{step}")
            func_opts(parser, sub_program=True)
            argument_docs = get_argument_docs_from_parser(parser)
        return argument_docs


    def write_step_doc(self, step):
        """
        folder: docs/folder/*.md
        """
        step_module = utils.find_step_module(self.assay, step)
        folder = step_module.__name__.split('.')[1]
        self.out_md_dict[step] = f'docs/{folder}/{step}.md'
        self.relative_md_path[step] = f'{folder}/{step}.md'

        class_docs = get_class_docs(step_module)
        argument_docs = self.get_argument_docs(step, step_module)

        with open(self.out_md_dict[step], 'w') as out_file:
            out_file.write(class_docs)
            out_file.write(argument_docs)

    def run(self):
        if self.release_bool:
            with open(MANUAL, 'a') as writer:
                writer.write(f'## {ASSAY_DICT[self.assay]}\n')            

        for step in self.steps:
            self.write_step_doc(step)
            if self.release_bool:
                self.write_step_in_manual(step)


    def write_step_in_manual(self, step):
        """
        - [mkref](rna/mkref.md)
        """
        if not step in PRE_PROCESSING_STEPS:
            with open(MANUAL, 'a') as writer:
                writer.write(f'- [{step}]({self.relative_md_path[step]})\n')


def main():
    cmd = (
        "set -eo pipefail;"
        f"rm -r {DOCS_DIR};"
        f"cp -r {TEMPLATE_DIR} {DOCS_DIR}"
    )
    os.system(cmd)

    for assay in ASSAY_DICT:
        docs_obj = Docs(assay)
        docs_obj.run()


if __name__ == "__main__":
    main()



import argparse

import celescope.tools.utils as utils
from celescope.__init__ import __VERSION__, ASSAY_DICT


class ArgFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
    pass


def main():
    """celescope cli
    """
    parser = argparse.ArgumentParser(description='CeleScope', formatter_class=ArgFormatter)
    parser.add_argument('-v', '--version', action='version', version=__VERSION__)
    subparsers = parser.add_subparsers()

    for assay in ASSAY_DICT:
        text = ASSAY_DICT[assay]
        subparser_1st = subparsers.add_parser(assay, description=text)
        # add 2ed subparser
        subparser_2ed = subparser_1st.add_subparsers()

        # import __STEPS__
        init_module = utils.find_assay_init(assay)
        __STEPS__ = init_module.__STEPS__

        for step in __STEPS__:
            # import function and opts
            step_module = utils.find_step_module(assay, step)
            func = getattr(step_module, step)
            func_opts = getattr(step_module, f"get_opts_{step}")
            parser_step = subparser_2ed.add_parser(step, formatter_class=ArgFormatter)
            func_opts(parser_step, sub_program=True)
            parser_step.set_defaults(func=func)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



#!/bin/env python
# coding=utf8

import os
import pandas as pd
import plotly
import plotly.graph_objects as go
from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils



class Replace_tsne(Step):
    """
    Features
    - Replace rate in each cluster
    - Top replace genes in each cluster

    Output
    - `{sample}.rep_in_tsne.txt` Replace rate in each cluster.
    - `{sample}.rep_in_tsne_top10` Top 10 replace genes in each cluster.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # input files
        self.sample = args.sample
        self.tsnefile = args.tsne 
        self.matfile = args.mat
        self.repfile = args.rep
        self.mincell = args.mincell
        self.topgene = args.topgene
        # output files
        self.outdot = os.path.join(self.outdir, self.sample+'.rep_in_tsne.txt')
        self.outtbl = os.path.join(self.outdir, self.sample+'.rep_in_tsne_top10.txt')

    @utils.add_log
    def run(self):
        # rep in cells in cluster
        self.dot_tsne(self.repfile,self.tsnefile,self.outdot)
        div_item = self.tsne_plot(self.outdot)
        # high rep gene in each cluster
        self.top_gene_cluster(self.matfile,self.tsnefile,self.outtbl,self.mincell,self.topgene)
        tbltxt = pd.read_csv(self.outtbl,header=0,sep="\t")
        tbldiv = self.tsne_table(tbltxt)

        # report
        self.report_prepare(div_item, tbldiv)
        self.clean_up()



    @utils.add_log
    def dot_tsne(self,repfile,tsnefile,outfile):
        cells = {}
        with open(repfile, 'r') as f:
            for i in f:
                ii = i.strip().split()
                cells[ii[0]] = ii[1]

        outf = open(outfile, 'w')
        outf.write("Cell\ttSNE_1\ttSNE_2\tCluster\tratio\n")
        with open(tsnefile, 'r') as f:
            f.readline()
            for i in f:
                ii = i.strip().split()
                if ii[0] in cells:
                    outl = '\t'.join(ii[0:4])+'\t'+cells[ii[0]]+'\n'
                else:
                    outl = '\t'.join(ii[0:4])+'\t0'+'\n'
                outf.write(outl)
        outf.close()

    @utils.add_log
    def tsne_plot(self,txt):
        df = pd.read_table(txt)
        df.sort_values(by="ratio")
        newtitle="t-SNE plot Colored by RNA Turn-over rate"

        fig = go.Figure()
        fig.add_trace(go.Scatter(x=df['tSNE_1'], y=df['tSNE_2'], mode='markers',
                            marker_opacity=0.9,marker_size=4,marker_color=df['ratio'],
                            marker_colorscale="PuBu", marker_showscale=True,
        ))
        fig.update_layout(height=600, width=600,title_text=newtitle)
        fig.update_layout(plot_bgcolor = '#FFFFFF')
        fig.update_xaxes(showgrid=False,linecolor='black', showline=True, ticks='outside',title_text='t-SNE1')
        fig.update_yaxes(showgrid=False,linecolor='black', showline=True, ticks='outside',title_text='t-SNE2')

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type='div')

        return div


    def tsne_table(self,txt):
        marker_gene_table = txt.to_html(
                escape=False,
                index=False,
                table_id='replacement_table_cluster',
                justify="center")

        return marker_gene_table


    def file_stat(self,infile,clu):
        clus = list(set(clu.values()))
        cluster = {}
        for c in clus:
            cluster[c] = {}
        fn = open(infile,"r")
        fnh = fn.readline().strip().split()
        fnh.insert(0,'')
        for i in fn:
            ii = i.strip().split()
            for j in range(1,len(ii)):
                if ii[j] == 'NA': continue
                if fnh[j] not in clu: continue
                if ii[0] not in cluster[clu[fnh[j]]]:
                    cluster[clu[fnh[j]]][ii[0]] = []
                cluster[clu[fnh[j]]][ii[0]].append(float(ii[j]))
        fn.close()
        return cluster

    def tsne_file(self,infile):
        clu = {}
        with open(infile) as f:
            f.readline()
            for i in f:
                ii = i.strip().split()
                clu[ii[0]] = ii[3]
        return clu


    @utils.add_log
    def top_gene_cluster(self,matrix,tsnefile,outfile,mincell=5,topgene=10):
        tsne = self.tsne_file(tsnefile)
        cluster = self.file_stat(matrix,tsne)

        w = open(outfile,'w')
        w.write("cluster\tgene\tTurn-over_rate\tcells\n")
        for c in cluster:
            tmp = {}
            for g in cluster[c]:
                gt = sum(cluster[c][g]) / len(cluster[c][g])
                tmp[g] = gt
            sorttmp = sorted(tmp.items(), key=lambda item:item[1], reverse=True)
            tmpn = 0
            for x in sorttmp:
                if len(cluster[c][x[0]]) < mincell: continue
                tmpn += 1
                if tmpn > topgene: break
                w.write('cluster'+c+'\t'+x[0]+'\t'+str(x[1])+'\t'+str(len(cluster[c][x[0]]))+'\n')
        w.close()


    def report_prepare(self,outdiv, outable):
        self.add_data_item(replace_tsne=outdiv)
        self.add_data_item(replace_tsne_table=outable)


@utils.add_log
def replace_tsne(args):

    step_name = "replace_tsne"
    replace_tsne_obj = Replace_tsne(args, step_name)
    replace_tsne_obj.run()

def get_opts_replace_tsne(parser, sub_program):
    if sub_program:
        parser.add_argument('--tsne', help='tsne file from analysis step', required=True)
        parser.add_argument('--mat', help='matrix replacement file, from replacement step', required=True)
        parser.add_argument('--rep', help='cell replacement file, from replacement step', required=True)
        parser.add_argument('--mincell', type=int, default=5, help='turn-over in at least cells, default 5')
        parser.add_argument('--topgene', type=int, default=10, help='show top N genes,default 10')
        parser = s_common(parser) 
    return parser




import configparser
import subprocess

import celescope.tools.utils as utils
from celescope.tools.mkref import Mkref, parse_genomeDir
from celescope.tools.mkref import get_opts_mkref as opts


def parse_genomeDir_rna(genomeDir):
    return parse_genomeDir(genomeDir, entrys=('fasta', 'gtf', 'mt_gene_list'))


class Mkref_rna(Mkref):
    """
    Features
    - Create a genome reference directory.

    Output

    - STAR genome index files

    - Genome refFlat file

    - Genome config file
    ```
    $ cat celescope_genome.config
    [genome]
    genome_name = Homo_sapiens_ensembl_99
    genome_type = rna
    fasta = Homo_sapiens.GRCh38.dna.primary_assembly.fa
    gtf = Homo_sapiens.GRCh38.99.gtf
    refflat = Homo_sapiens_ensembl_99.refFlat
    ```
    """

    def __init__(self, genome_type, args):
        Mkref.__init__(self, genome_type, args)
        self.fasta = args.fasta
        self.gtf = args.gtf
        self.mt_gene_list = args.mt_gene_list

        # out file
        self.refflat = f'{self.genome_name}.refFlat'

    @utils.add_log
    def build_star_index(self):
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--genomeDir {self.genomeDir} \\\n'
            f'--genomeFastaFiles {self.fasta} \\\n'
            f'--sjdbGTFfile {self.gtf} \\\n'
            f'--sjdbOverhang 100 \\\n'
        )
        Mkref_rna.build_star_index.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def write_config(self):
        config = configparser.ConfigParser()
        config['genome'] = {}
        genome = config['genome']
        genome['genome_name'] = self.genome_name
        genome['genome_type'] = self.genome_type
        genome['fasta'] = self.fasta
        genome['gtf'] = self.gtf
        genome['refFlat'] = self.refflat
        genome['mt_gene_list'] = self.mt_gene_list
        with open(self.config_file, 'w') as config_handle:
            config.write(config_handle)

    @utils.add_log
    def build_refflat(self):
        cmd = (
            'gtfToGenePred -genePredExt -geneNameAsName2 \\\n'
            f'{self.gtf} /dev/stdout | \\\n'
            'awk \'{print $12"\\t"$1"\\t"$2"\\t"$3"\\t"$4"\\t"$5"\\t"$6"\\t"$7"\\t"$8"\\t"$9"\\t"$10}\' \\\n'
            f'> {self.refflat} \\\n'
        )
        Mkref_rna.build_refflat.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        if not self.dry_run:
            self.build_refflat()
            self.build_star_index()
        self.write_config()


def mkref(args):
    genome_type = 'rna'
    runner = Mkref_rna(genome_type, args)
    runner.run()


def get_opts_mkref(parser, sub_program):
    opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--fasta",
            help="Required. Genome fasta file. Must be relative file path to genomeDir.",
            required=True
        )
        parser.add_argument(
            "--gtf",
            help="Required. Genome gtf file. Must be relative file path to genomeDir.",
            required=True
        )
        parser.add_argument(
            "--mt_gene_list",
            help="""Mitochondria gene list file. Must be relative file path to genomeDir.
It is a plain text file with one gene per line. 
If not provided, will use `MT-` and `mt-` to determine mitochondria genes.""",
            default="None"
        )



#!/bin/env python
# coding=utf8

import os
import sys
import subprocess
import pandas as pd
import pysam
from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils

toolsdir = os.path.dirname(__file__)


class Replacement(Step):
    """
    Features
    - Computes the replacement rates in each cell and gene.
    - Boxplots for rates distribution.

    Output
    - `{sample}.TC_matrix.rds` New and old info for each barcode/gene/umi.
    - `{sample}.new_matrix.tsv.gz` New RNA matrix.
    - `{sample}.old_matrix.tsv.gz` Old RNA matrix.
    - `{sample}.fraction_of_newRNA_per_cell.txt` Fraction of new RNA of each cell.
    - `{sample}.fraction_of_newRNA_per_gene.txt` Fraction of new RNA of each gene.
    - `{sample}.fraction_of_newRNA_matrix.txt` Fraction of new RNA of each cell and gene.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # input files
        self.outdir = args.outdir
        self.sample = args.sample
        self.bam_file = args.bam
        self.snp_file = args.bg
        self.bg_cov = args.bg_cov
        self.cell_keep = args.cell_keep
        # output files
        self.outread = os.path.join(self.outdir, self.sample+'.corrected_gene_cell_UMI_read.txt')
        self.outrds = os.path.join(self.outdir, self.sample+'.TC_matrix.rds')
        self.outpre = os.path.join(self.outdir,self.sample)
    
    @utils.add_log
    def run(self):
        # get backgroud snp
        bg = self.background_snp(self.snp_file,self.bg_cov)
        # get reads with TC
        self.extract_dem(self.bam_file,self.outread,bg)
        # run_R
        self.generate_TC_matrix(self.outread, self.outrds, self.cell_keep)

        # split to New and Old Matrix
        totMat = self.outrds+'.tsv'
        new_mat = self.outpre+'.new_matrix.tsv'
        old_mat = self.outpre+'.old_matrix.tsv'
        con_mat = self.outpre+'.NvsO_matrix.tsv'
        self.split_matrix(totMat,self.outpre)
        
        # replacement stat
        self.replacment_stat(con_mat,self.outpre)
        # plot
        div_item = self.replacment_plot(self.outpre)

        # report
        self.report_prepare(div_item)
        self.clean_up()

        # clean
        cmd=['rm', self.outread]
        self.run_cmd(cmd)
        cmd=['rm', self.outrds+'.tsv']
        self.run_cmd(cmd)
        cmd=['rm', con_mat]
        self.run_cmd(cmd)
        cmd=['gzip', new_mat]
        self.run_cmd(cmd)
        cmd=['gzip', old_mat]
        self.run_cmd(cmd)

    def run_cmd(self,cmd):
        subprocess.call(' '.join(cmd),shell=True)

    @utils.add_log
    def extract_dem(self,bam,outfile,bg):
        bamfile = pysam.AlignmentFile(bam, 'rb')
        countdict = {}
        for read in bamfile.fetch():
            try:
                chro = read.reference_name
                cb = read.get_tag('CB')
                ub = read.get_tag('UB')
                if not read.has_tag('GN'): continue
                gene = read.get_tag('GN')

                if read.get_tag('ST') == '+':
                    stag = read.get_tag('TL')
                else:
                    stag = read.get_tag('AL')
                if len(stag)==1 and stag[0]==0:
                    gene += '--T'
                else:
                    fcount = 0
                    for si in range(0,len(stag)):
                        pos = chro + '_' + str(stag[si])
                        if pos in bg:
                            fcount += 1
                    if fcount == len(stag):
                        gene += '--T'
                    else:
                        gene += '--C'
                
                readinfo = '\t'.join([gene,cb,ub])
                if readinfo not in countdict:
                    countdict[readinfo] = 1
                else:
                    countdict[readinfo] += 1

            except (ValueError,KeyError):
                continue
        bamfile.close()

        out1 = open(outfile,'w')
        for rid in countdict:
            out1.write(rid+'\t'+str(countdict[rid])+'\n')
        out1.close()

    @utils.add_log
    def background_snp(self,bgfile,cov=1):
        outdict = {}
        if bgfile.endswith('.csv'):
            with open(bgfile) as f:
                f.readline()
                for i in f:
                    ii = i.strip().split(',')
                    if int(ii[2])<cov: continue
                    chr_pos = ii[4]+'_'+ii[1]
                    outdict[chr_pos] = 1
        elif bgfile.endswith('.vcf'):
            bcf_in = pysam.VariantFile(bgfile)
            for rec in bcf_in.fetch():
                try:
                    chrom, pos = rec.chrom, rec.pos
                    chr_pos = chrom+'_'+str(pos)
                    outdict[chr_pos] = 1
                except (ValueError,KeyError):
                    continue
            bcf_in.close()

        else:
            try:
                sys.exit(1)
            except SystemExit:
                print('Background snp file format cannot be recognized! Only csv or vcf format.')
            finally:
                print('Background snp file format cannot be recognized! Only csv or vcf format.')
        return outdict


    @utils.add_log
    def generate_TC_matrix(self,read, outrds, cell=100000):
        app = toolsdir + "/Generate_T_C_matrix.R"
        cmd = (
            f'Rscript {app} {read} {cell} {outrds}'
        )
        os.system(cmd)

    @utils.add_log
    def split_matrix(self,mat,outpre):
        outnew = open(outpre+'.new_matrix.tsv', 'w')
        outold = open(outpre+'.old_matrix.tsv', 'w')
        con_mat = open(outpre+'.NvsO_matrix.tsv', 'w')
        infile = open(mat, 'r')

        tmph = infile.readline().strip().split()
        fill_na = ['0'] * len(tmph)
        tmph.insert( 0, '')
        outnew.write('\t'.join(tmph)+'\n')
        outold.write('\t'.join(tmph)+'\n')
        con_mat.write('\t'.join(tmph)+'\n')

        genes = {}
        for i in infile:
            ii = i.strip().split()
            gt = ii[0].split('--')
            ii[0] = gt[0]
            if gt[0] not in genes:
                genes[gt[0]] = [0,[],[]]
            if gt[1] == 'C':
                genes[gt[0]][0] += 1
                genes[gt[0]][1] = ii[1:]
                outnew.write('\t'.join(ii)+'\n')
            elif gt[1] == 'T':
                genes[gt[0]][0] += 2
                genes[gt[0]][2] = ii[1:]
                outold.write('\t'.join(ii)+'\n')
        
        for gi in genes:
            con_mat.write(gi)
            if genes[gi][0]==3:
                for ci in range(len(genes[gi][1])):
                    con_mat.write('\t'+genes[gi][1][ci]+':'+genes[gi][2][ci])
                con_mat.write('\n')
            elif genes[gi][0]==1:
                outold.write(gi+'\t'+'\t'.join(fill_na)+'\n')
                for ci in range(len(genes[gi][1])):
                    con_mat.write('\t'+genes[gi][1][ci]+':'+'0')
                con_mat.write('\n')
            elif genes[gi][0]==2:
                outnew.write(gi+'\t'+'\t'.join(fill_na)+'\n')
                for ci in range(len(genes[gi][2])):
                    con_mat.write('\t'+'0'+':'+genes[gi][2][ci])
                con_mat.write('\n')

        outnew.close()
        outold.close()
        con_mat.close()
        infile.close()


    @utils.add_log
    def replacment_stat(self,inmat,outpre,mincell=10,mingene=10):

        outcell = open(outpre+'.fraction_of_newRNA_per_cell.txt', 'w')
        outgene = open(outpre+'.fraction_of_newRNA_per_gene.txt', 'w')
        outmat = open(outpre+'.fraction_of_newRNA_matrix.txt', 'w')

        cells = {}
        genes = {}
        mats = {}
        with open(inmat) as f:
            hh = f.readline().strip().split()
            hh.insert(0,'')
            outmat.write('\t'.join(hh)+'\n')
            for h in hh[1:]:
                cells[h] = [[],[]]
            for i in f:
                ii = i.strip().split()
                mats[ii[0]] = []
                genes[ii[0]] = [[],[]]
                for xi in range(1,len(ii)):
                    xx = [int(x) for x in ii[xi].split(':')]                 
                    if sum(xx) == 0:
                        tmpf = 'NA'                   
                    else:
                        tmpf = float(xx[0])/(xx[0]+xx[1])
                    mats[ii[0]].append(str(tmpf))
                    if sum(xx)<2:
                        continue                
                    cells[hh[xi]][0].append(float(xx[0]))
                    cells[hh[xi]][1].append(xx[1])
                    genes[ii[0]][0].append(float(xx[0]))
                    genes[ii[0]][1].append(xx[1])

        for ci in cells:
            if len(cells[ci][0])<mincell: continue
            cfloat = sum(cells[ci][0])/(sum(cells[ci][0])+sum(cells[ci][1]))
            outcell.write(ci+'\t'+str(cfloat)+'\n')
        
        for gi in genes:
            if len(genes[gi][0])<mingene: continue
            gfloat = sum(genes[gi][0])/(sum(genes[gi][0])+sum(genes[gi][1]))
            outgene.write(gi+'\t'+str(gfloat)+'\n')

        for mi in mats:
            outmat.write(mi+'\t'+'\t'.join(mats[mi])+'\n')

        outcell.close()
        outgene.close()
        outmat.close()


    @utils.add_log
    def replacment_plot(self,sample):
        import plotly
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots

        outpre = os.path.basename(sample)
        df1 = pd.read_table(sample+'.fraction_of_newRNA_per_gene.txt', header=None)
        df1.columns = ['gene', 'per']
        df = pd.read_table(sample+'.fraction_of_newRNA_per_cell.txt', header=None)
        df.columns = ['gene', 'per']

        fig = make_subplots(rows=1, cols=2)
        fig.add_trace(
            go.Violin(y=df1['per'], box_visible=True, line_color='black',
                meanline_visible=True, fillcolor='#1f77b4', opacity=0.6, x0=outpre) ,
            row=1, col=1
        )
        fig.add_trace(
            go.Violin(y=df['per'], box_visible=True, line_color='black',
                meanline_visible=True, fillcolor='#ff7f0e', opacity=0.6, x0=outpre) ,
            row=1, col=2
        )

        fig.update_layout(yaxis_zeroline=True,  showlegend=False)
        fig.update_layout(plot_bgcolor = '#FFFFFF')
        fig.update_xaxes(showgrid=False, linecolor='black', showline=True, ticks=None)
        fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside',title_text="Fraction new RNA (per gene)",row=1, col=1, rangemode="tozero")
        fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside',title_text="Fraction new RNA (per cell)",row=1, col=2, rangemode="tozero")

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type='div')

        return div

    def report_prepare(self,outdiv):
        self.add_data_item(replacement=outdiv)


@utils.add_log
def replacement(args):

    step_name = "replacement"
    replacement_obj = Replacement(args, step_name)
    replacement_obj.run()


def get_opts_replacement(parser, sub_program):
    parser.add_argument('--bg_cov', type=int, default=1, help='background snp depth filter, lower than bg_cov will be discarded. Only valid in csv format')
    if sub_program:
        parser.add_argument('--bam', help='bam file from conversion step', required=True)
        parser.add_argument('--bg', help='background snp file, csv or vcf format', required=True)
        parser.add_argument('--cell_keep', type=int, default=100000, help='filter cell')
        parser.add_argument('--min_cell', type=int, default=10, help='a gene expressed in at least cells, default 10')
        parser.add_argument('--min_gene', type=int, default=10, help='at least gene num in a cell, default 10')
        parser = s_common(parser) 
    return parser





import os
import subprocess
from concurrent.futures import ProcessPoolExecutor
from os import listdir
from os.path import isfile, join

from celescope.tools.step import Step, s_common
from celescope.tools.utils import add_log

TRACER_PATH = '/SGRNJ/Public/Software/tracer/tracer'
CONF_PATH = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/tcr_fl/20201103/tracer_SGR.conf'
CONDA = 'vdjpuzzle1'
CONDA_SUB = 'celescope_tracer'


def tracer(fq, outdir):
    prefix = os.path.basename(fq).strip('.fq')
    cmd = (
        f'source activate {CONDA}; '
        f'{TRACER_PATH} assemble '
        f'--fragment_length 150 '
        f'--fragment_sd 5 '
        f'--single_end '
        f'--species Hsap '
        f'-c {CONF_PATH} '
        f'{fq} '
        f'{prefix} '
        f'{outdir}/tracer '
    )
    subprocess.check_call(cmd, shell=True)


class Assemble_TCR(Step):
    def __init__(self, args, step):
        Step.__init__(self, args, step)
        self.fastq_dir = args.fastq_dir

    @add_log
    def tracer_summarise(self):
        tracer_outdir = f'{self.outdir}/tracer'
        cmd = (
            f'source activate {CONDA_SUB}; '
            f'{TRACER_PATH} summarise '
            f'-c {CONF_PATH} '
            f'{tracer_outdir} '
        )
        Assemble_TCR.tracer_summarise.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @add_log
    def run(self):
        fqs = [join(self.fastq_dir, f) for f in listdir(self.fastq_dir) if isfile(join(self.fastq_dir, f))]
        outdirs = [self.outdir] * len(fqs)
        if not os.path.exists(f'{self.outdir}/tracer'):
            os.makedirs(f'{self.outdir}/tracer')

        all_res = []
        with ProcessPoolExecutor(self.thread) as pool:
            for res in pool.map(tracer, fqs, outdirs):
                all_res.append(res)
        self.tracer_summarise()
        self.clean_up()


def assemble(args):
    step_name = "assemble"
    runner = Assemble_TCR(args, step_name)
    runner.run()


def get_opts_assemble(parser, sub_program):
    s_common(parser)
    if sub_program:
        parser.add_argument("--fastq_dir", required=True)



from celescope.rna.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_rna(Multi):
    """
    Usage
    ```
        multi_rna\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --thread 8\\
        --mod shell
    ```

    If Single nuclei RNA-Seq is used, you need to add `--gtf_type gene` to include reads mapped to 
    intronic regions.
    """


def main():
    multi = Multi_rna(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()




from celescope.mut.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_mut(Multi):

    def mapping_mut(self, sample):
        step = 'mapping_mut'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd = (
            f'{self.__APP__} '
            f'{self.__ASSAY__} '
            f'{step } '
            f'--outdir {self.outdir_dic[sample][step]} '
            f'--sample {sample} '
            f'--assay {self.__ASSAY__} '
            f'--fq {fq} '
            f'--thread {self.thread} '
            f'--indel_genomeDir {self.indel_genomeDir} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_mut(self, sample):
        step = 'count_mut'
        bam = f'{self.outdir_dic[sample]["mapping_mut"]}/{sample}_Aligned.sortedByCoord.out.bam'
        cmd = (
            f'{self.__APP__} '
            f'{self.__ASSAY__} '
            f'{step } '
            f'--outdir {self.outdir_dic[sample][step]} '
            f'--sample {sample} '
            f'--assay {self.__ASSAY__} '
            f'--bam {bam} '
            f'--mut_file {self.mut_file} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--shift_base {self.shift_base} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=1)


def main():
    multi = Multi_mut(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()




import numpy as np
import pysam
import sys

import celescope.tools.utils as utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT


class Target_metrics(Step):
    """
    Features
    - Filter bam file
        - Filter reads that are not cell-associated.
        - Filter reads that are not mapped to target genes. 

    - Collect enrichment metrics.

    Output
    - `filtered.bam` BAM file after filtering.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.match_barcode, self.n_cell = utils.read_barcode_file(args.match_dir)
        self.match_barcode = set(self.match_barcode)
        
        if (self.assay == "snp" and args.panel != '') :
            self.gene_list = utils.get_gene_region_from_bed(args.panel)[0]
            self.n_gene = len(self.gene_list)
        else:
            self.gene_list, self.n_gene = utils.read_one_col(args.gene_list)

        if not self.gene_list:
            sys.exit("You must provide either --panel or --gene_list!")            
        
        self.count_dict = utils.genDict(dim=3, valType=int)

        self.add_metric(
            name="Number of Target Genes",
            value=self.n_gene,
        )
        self.add_metric(
            name="Number of Cells",
            value=self.n_cell,
        )

        # out file
        self.out_bam_file = f'{self.out_prefix}_filtered.bam'
        self.out_bam_file_sorted = f'{self.out_prefix}_filtered_sorted.bam'

    @utils.add_log
    def read_bam_write_filtered(self):
        with pysam.AlignmentFile(self.args.bam, "rb") as reader:
            with pysam.AlignmentFile(self.out_bam_file, "wb", header=reader.header) as writer:
                for record in reader:
                    try:
                        gene_name = record.get_tag('GN')
                    except KeyError:
                        continue
                    # compatible with 10X bam
                    try:
                        barcode = record.get_tag('CB')
                        UMI = record.get_tag('UB')
                    except KeyError:
                        continue
                    if barcode in self.match_barcode and gene_name in self.gene_list:
                        writer.write(record)
                    self.count_dict[barcode][gene_name][UMI] += 1

    @utils.add_log
    def parse_count_dict_add_metrics(self):
        total_reads = 0
        enriched_reads = 0
        enriched_reads_in_cells = 0
        enriched_reads_per_cell_list = []

        for barcode in self.count_dict:
            cell_enriched_read = 0
            for gene_name in self.count_dict[barcode]:
                gene_read = sum(self.count_dict[barcode][gene_name].values())
                total_reads += gene_read
                if gene_name in self.gene_list:
                    enriched_reads += gene_read
                    if barcode in self.match_barcode:
                        enriched_reads_in_cells += gene_read
                        cell_enriched_read += gene_read
            if barcode in self.match_barcode:
                enriched_reads_per_cell_list.append(cell_enriched_read)

        if self.debug:
            self.parse_count_dict_add_metrics.logger.debug(
                f'enriched_reads_per_cell_list: '
                f'{sorted(enriched_reads_per_cell_list)}'
                f'len: {len(enriched_reads_per_cell_list)}'
            )
        
        n_valid_cell = len([cell for cell in enriched_reads_per_cell_list if cell > 0])
        self.add_metric(
            name="Number of Valid Cells",
            value=n_valid_cell,
            total=self.n_cell,
        )        
        self.add_metric(
            name="Enriched Reads",
            value=enriched_reads,
            total=total_reads,
        )
        self.add_metric(
            name="Enriched Reads in Cells",
            value=enriched_reads_in_cells,
            total=total_reads,
        )
        self.add_metric(
            name="Median Enriched Reads per Valid Cell",
            value=np.median(enriched_reads_per_cell_list),
        )

    def run(self):
        self.read_bam_write_filtered()
        self.parse_count_dict_add_metrics()
        utils.sort_bam(
            self.out_bam_file,
            self.out_bam_file_sorted,
            threads=self.thread,
        )
        utils.index_bam(self.out_bam_file_sorted)
        self.clean_up()


@utils.add_log
def target_metrics(args):
    step_name = "target_metrics"
    runner = Target_metrics(args, step_name)
    runner.run()


def get_opts_target_metrics(parser, sub_program):
    parser.add_argument("--gene_list", help=HELP_DICT['gene_list'])
    parser.add_argument("--panel",help = "The prefix of bed file, such as `lung_1`.",default = '')
    if sub_program:
        parser.add_argument("--bam", help='Input bam file', required=True)
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser = s_common(parser)



import subprocess

import pandas as pd

import celescope.tools.utils as utils
from celescope.__init__ import ROOT_PATH
from celescope.rna.mkref import parse_genomeDir_rna
from celescope.tools.step import Step


class AnalysisMixin():
    """
    mixin class for analysis
    child class must inherite Step class
    """

    def __init__(self, args):
        self.args = args
        self.match_dir = None
        if hasattr(args, "match_dir") and args.match_dir:
            self.match_dir = args.match_dir
            self.read_match_dir()
        elif hasattr(args, "tsne_file") and args.tsne_file:
            tsne_df_file = args.tsne_file
            self.tsne_df = pd.read_csv(tsne_df_file, sep="\t")
            self.tsne_df.rename(columns={"Unnamed: 0": "barcode"}, inplace=True)
            marker_df_file = args.marker_file
            self.marker_df = pd.read_csv(marker_df_file, sep="\t")
        else:
            self.match_dir = args.outdir + "/../"  # use self

    @utils.add_log
    def seurat(self, matrix_file, save_rds, genomeDir):
        app = ROOT_PATH + "/tools/run_analysis.R"
        genome = parse_genomeDir_rna(genomeDir)
        mt_gene_list = genome['mt_gene_list']
        cmd = (
            f'Rscript {app} '
            f'--sample {self.sample} '
            f'--outdir {self.outdir} '
            f'--matrix_file {matrix_file} '
            f'--mt_gene_list {mt_gene_list} '
            f'--save_rds {save_rds}'
        )
        self.seurat.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def auto_assign(self, type_marker_tsv):
        rds = f'{self.outdir}/{self.sample}.rds'
        app = ROOT_PATH + "/tools/auto_assign.R"
        cmd = (
            f'Rscript {app} '
            f'--rds {rds} '
            f'--type_marker_tsv {type_marker_tsv} '
            f'--outdir {self.outdir} '
            f'--sample {self.sample} '
        )
        self.auto_assign.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    def get_cluster_tsne(colname, tsne_df, show_colname=True):
        """
        tSNE_1	tSNE_2	cluster Gene_Counts
        return data list
        """

        sum_df = tsne_df.groupby([colname]).agg("count").iloc[:, 0]
        percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
        res = []
        for cluster in sorted(tsne_df[colname].unique()):
            sub_df = tsne_df[tsne_df[colname] == cluster]
            if show_colname:
                name = f"{colname} {cluster}({percent_df[cluster]}%)"
            else:
                name = f"{cluster}({percent_df[cluster]}%)"
            tSNE_1 = list(sub_df.tSNE_1)
            tSNE_2 = list(sub_df.tSNE_2)
            res.append({"name": name, "tSNE_1": tSNE_1, "tSNE_2": tSNE_2})
        return res

    @staticmethod
    def get_table_dict(marker_table):
        table_dict = Step.get_table(
            title='Marker Genes by Cluster',
            table_id='marker_gene_table',
            df_table=marker_table,
        )
        return table_dict

    def get_marker_table(self):
        """
        return html code
        """

        avg_logfc_col = "avg_log2FC"  # seurat 4
        if "avg_logFC" in self.marker_df.columns:  # seurat 2.3.4
            avg_logfc_col = "avg_logFC"
        marker_df = self.marker_df.loc[:,
                                       ["cluster", "gene", avg_logfc_col, "pct.1", "pct.2", "p_val_adj"]
                                       ]
        marker_df["cluster"] = marker_df["cluster"].apply(lambda x: f"cluster {x}")

        return marker_df

    def get_gene_tsne(self):
        """
        return data dic
        """
        tsne_df = self.tsne_df
        tSNE_1 = list(tsne_df.tSNE_1)
        tSNE_2 = list(tsne_df.tSNE_2)
        Gene_Counts = list(tsne_df.Gene_Counts)
        res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "Gene_Counts": Gene_Counts}
        return res

    def read_match_dir(self):
        """
        if match_dir is not self, should read match_dir at init
        if it is self, read at run_analysis - need to run seurat first
        """
        if self.match_dir:
            match_dict = utils.parse_match_dir(self.match_dir)
            tsne_df_file = match_dict['tsne_coord']
            self.tsne_df = pd.read_csv(tsne_df_file, sep="\t")
            self.tsne_df.rename(columns={"Unnamed: 0": "barcode"}, inplace=True)
            marker_df_file = match_dict['markers']
            self.marker_df = pd.read_csv(marker_df_file, sep="\t")

    def run_analysis(self):
        self.read_match_dir()
        self.cluster_tsne = self.get_cluster_tsne(colname='cluster', tsne_df=self.tsne_df)
        self.gene_tsne = self.get_gene_tsne()
        marker_table = self.get_marker_table()
        self.table_dict = self.get_table_dict(marker_table)



import subprocess

import pandas as pd
import pysam

import celescope.tools.utils as utils
from celescope.__init__ import ROOT_PATH, HELP_DICT
from celescope.fusion.mkref import parse_genomeDir_fusion
from celescope.tools.step import Step, s_common


class CountFusion(Step):
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        self.flanking_base = int(args.flanking_base)
        self.UMI_min = int(args.UMI_min)
        self.match_dir = args.match_dir
        self.fusion_genomeDir = args.fusion_genomeDir
        self.fusion_pos_file = parse_genomeDir_fusion(self.fusion_genomeDir)['fusion_pos']
        self.bam = args.bam

        # set
        self.pos_dic = self.read_pos()
        self.match_barcode, self.n_match_barcode = utils.read_barcode_file(self.match_dir)
        # tsne
        match_tsne_file = utils.parse_match_dir(self.match_dir)["tsne_coord"]
        self.df_tsne = pd.read_csv(match_tsne_file, sep="\t", index_col=0)

        # out
        self.fusion_bam = self.out_prefix + "_fusion.bam"
        self.out_read_count_file = self.out_prefix + "_fusion_read_count.tsv"
        self.out_umi_count_file = self.out_prefix + "_fusion_UMI_count.tsv"
        self.out_barcode_count_file = self.out_prefix + "_fusion_barcode_count.tsv"
        self.out_tsne_file = self.out_prefix + "_fusion_tsne.tsv"

    def read_pos(self):
        """
        pos_dic
            key: sequence name in fusion.fasta
            value: end postion of the first gene(1-based).
        """
        pos_dic = {}
        df = pd.read_csv(self.fusion_pos_file, sep="\t")
        for name, pos in zip(df.iloc[:, 0], df.iloc[:, 1]):
            pos_dic[name] = pos
        return pos_dic

    def count_fusion(self):
        """
        find valid fusion reads
            1. flank the fusion position
            2. match barcode
        """
        count_dic = utils.genDict(dim=3)

        with pysam.AlignmentFile(self.bam, "rb") as bam:
            header = bam.header
            with pysam.AlignmentFile(self.fusion_bam, "wb", header=header) as fusion_bam:
                for name in self.pos_dic:
                    pos = self.pos_dic[name]
                    left = pos - self.flanking_base
                    right = pos + self.flanking_base
                    for read in bam.fetch(
                        reference=name,
                        start=left,
                        end=right,
                    ):
                        left_bases = read.get_overlap(left, pos)
                        right_bases = read.get_overlap(pos, right)
                        if left_bases < self.flanking_base or right_bases <  self.flanking_base:
                            continue
                        attr = read.query_name.split("_")
                        barcode = attr[0]
                        umi = attr[1]
                        if barcode in self.match_barcode:
                            fusion_bam.write(read)
                            count_dic[barcode][name][umi] += 1

        # write dic to pandas df
        rows = []
        for barcode in count_dic:
            for name in count_dic[barcode]:
                for umi in count_dic[barcode][name]:
                    rows.append([barcode, name, umi, count_dic[barcode][name][umi]])
        if not rows:
            count_fusion.logger.warning("***** NO FUSION FOUND! *****")
            return
        df_read = pd.DataFrame(rows)
        df_read.rename(
            columns={0: "barcode", 1: "name", 2: "UMI", 3: "read_count"}, inplace=True
        )
        df_read.to_csv(self.out_read_count_file, sep="\t", index=False)

        df_umi = df_read.groupby(["barcode", "name"]).agg({"UMI": "count"})
        df_umi = df_umi[df_umi["UMI"] >= self.UMI_min]
        df_umi.to_csv(self.out_umi_count_file, sep="\t")

        df_umi.reset_index(inplace=True)
        df_barcode = df_umi.groupby(["name"]).agg({"barcode": "count"})
        # add zero count name
        for name in self.pos_dic:
            if not name in df_barcode.index:
                new_row = pd.Series(data={"barcode": 0}, name=name)
                df_barcode = df_barcode.append(new_row, ignore_index=False)
        df_barcode["percent"] = df_barcode["barcode"] / self.n_match_barcode
        df_barcode.to_csv(self.out_barcode_count_file, sep="\t")

        df_pivot = df_umi.pivot(index="barcode", columns="name", values="UMI")
        df_pivot.fillna(0, inplace=True)
        df_tsne_fusion = pd.merge(
            self.df_tsne, df_pivot, right_index=True, left_index=True, how="left"
        )
        df_tsne_fusion.fillna(0, inplace=True)
        df_tsne_fusion.to_csv(self.out_tsne_file, sep="\t")

        self.plot_fusion()

    @utils.add_log
    def plot_fusion(self):
        app = f'{ROOT_PATH}/fusion/plot_fusion.R'
        cmd = f"Rscript {app} --tsne_fusion {self.out_tsne_file} --outdir {self.outdir}"
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.count_fusion()
        self.clean_up()


@utils.add_log
def count_fusion(args):
    step_name = "count_fusion"
    runner = CountFusion(args, step_name)
    runner.run()


def get_opts_count_fusion(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument("--bam", help='STAR bam file.', required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'], required=True)
    parser.add_argument('--fusion_genomeDir', help='Fusion genome directory.', required=True)
    parser.add_argument(
        "--flanking_base", 
        help="Number of bases flanking the fusion position.",
        default=5)
    parser.add_argument(
        "--UMI_min",
        help="Minimum number of fusion UMI to consider a cell as a cell with fusion event.",
        default=1)



import configparser
import subprocess

import celescope.tools.utils as utils
from celescope.tools.mkref import Mkref
from celescope.tools.mkref import get_opts_mkref as opts
from celescope.tools.mkref import parse_genomeDir


def parse_genomeDir_fusion(genomeDir):
    return parse_genomeDir(genomeDir, entrys=('fasta', 'fusion_pos'))


class Mkref_fusion(Mkref):
    """
    Features
    - Create a fusion genome directory.

    Output

    - STAR genome index files
    - Genome config file
    
    Usage
    ```
    celescope fusion mkref \\
    --genome_name {genome_name} \\
    --fasta fusion.fasta \\
    --fusion_pos fusion_pos.txt \\
    --genomeSAindexNbases 4
    ```
    """
    def __init__(self, genome_type, args):
        Mkref.__init__(self, genome_type, args)
        self.fasta = args.fasta
        self.genomeSAindexNbases = args.genomeSAindexNbases
        self.fusion_pos = args.fusion_pos

    @utils.add_log
    def build_star_index(self):
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--genomeDir {self.genomeDir} \\\n'
            f'--genomeFastaFiles {self.fasta} \\\n'
            f'--genomeSAindexNbases {self.genomeSAindexNbases} \\\n'
        )
        Mkref_fusion.build_star_index.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def write_config(self):
        config = configparser.ConfigParser()
        config['genome'] = {}
        genome = config['genome']
        genome['genome_name'] = self.genome_name
        genome['genome_type'] = self.genome_type
        genome['fasta'] = self.fasta
        genome['fusion_pos'] = self.fusion_pos
        genome['genomeSAindexNbases'] = self.genomeSAindexNbases
        with open(self.config_file, 'w') as config_handle:
            config.write(config_handle)

    def run(self):
        if not self.dry_run:
            self.build_star_index()
        self.write_config()


def mkref(args):
    genome_type = 'fusion'
    runner = Mkref_fusion(genome_type, args)
    runner.run()


def get_opts_mkref(parser, sub_program):
    opts(parser, sub_program)
    if sub_program:
        parser.add_argument("--fasta", help="Fusion fasta file.", required=True)
        parser.add_argument(
            "--fusion_pos",
            help="""
fusion position file. A two column tab-delimited text file with header.
"pos" is the end postion of the first gene(1-based).
e.g.  
name\tpos  
PML_3\t183  
PML_4\t254  
PML_5\t326  
PML_6\t204   
""",
            required=True,)
        parser.add_argument("--genomeSAindexNbases", help="STAR genomeSAindexNbases", default=4)



import celescope.tools.utils as utils
from celescope.tools.star_mixin import StarMixin, get_opts_star_mixin
from celescope.tools.step import Step


class StarVirus(Step, StarMixin):
    """
    star virus class
    """

    def __init__(self, args, step_name):
        # add genomeDir
        args.genomeDir = args.virus_genomeDir

        Step.__init__(self, args, step_name)
        StarMixin.__init__(self, args, add_prefix='virus')

    def run(self):
        self.run_star()
        self.clean_up()


@utils.add_log
def star_virus(args):
    step_name = "star_virus"
    runner = StarVirus(args, step_name)
    runner.run()


def get_opts_star_virus(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)
    parser.add_argument('--virus_genomeDir', help='virus genome dir', required=True)



import re
import subprocess

import celescope.tools.utils as utils
from celescope.tools.mkref import parse_genomeDir
from celescope.tools.step import s_common
from celescope.__init__ import HELP_DICT

class StarMixin():
    """
    Mixin class for STAR
    """

    def __init__(self, args, add_prefix=None):
        self.fq = args.fq
        self.genomeDir = args.genomeDir
        self.out_unmapped = args.out_unmapped
        self.debug = args.debug
        self.outFilterMatchNmin = int(args.outFilterMatchNmin)
        self.multi_max = int(args.outFilterMultimapNmax)
        self.STAR_param = args.STAR_param
        self.consensus_fq = args.consensus_fq

        # parse
        self.genome = parse_genomeDir(self.genomeDir)
        self.stat_prefix = 'Reads'
        if self.consensus_fq:
            self.stat_prefix = 'UMIs'

        # out
        self.outPrefix = f'{self.outdir}/{self.sample}_'
        if add_prefix:
            self.outPrefix += add_prefix + '_'
        self.STAR_map_log = f'{self.outPrefix}Log.final.out'
        self.unsort_STAR_bam = f'{self.outPrefix}Aligned.out.bam'
        self.STAR_bam = f'{self.outPrefix}Aligned.sortedByCoord.out.bam'

    @utils.add_log
    def STAR(self):
        cmd = [
            'STAR',
            '--runThreadN', str(self.thread),
            '--genomeDir', self.genomeDir,
            '--readFilesIn', self.fq,
            '--outFilterMultimapNmax', str(self.multi_max),
            '--outFileNamePrefix', self.outPrefix,
            '--outSAMtype', 'BAM', 'Unsorted',  # controls sort by Coordinate or not
            '--outFilterMatchNmin', str(self.outFilterMatchNmin)
        ]
        if self.out_unmapped:
            cmd += ['--outReadsUnmapped', 'Fastx']
        if self.fq[-3:] == ".gz":
            cmd += ['--readFilesCommand', 'zcat']
        cmd = ' '.join(cmd)
        if self.STAR_param:
            cmd += (" " + self.STAR_param)
        StarMixin.STAR.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run_star(self):
        self.STAR()
        self.get_star_metrics()
        self.sort_bam()
        self.index_bam()

    @utils.add_log
    def sort_bam(self):
        utils.sort_bam(
            self.unsort_STAR_bam,
            self.STAR_bam,
            threads=self.thread,
        )

    @utils.add_log
    def index_bam(self):
        utils.index_bam(self.STAR_bam)

    def get_star_metrics(self):
        """
        step metrics
        """

        with open(self.STAR_map_log, 'r') as map_log:
            # number amd percent
            unique_reads_list = []
            multi_reads_list = []
            for line in map_log:
                if line.strip() == '':
                    continue
                if re.search(r'Uniquely mapped reads', line):
                    unique_reads_list.append(line.strip().split()[-1])
                if re.search(r'of reads mapped to too many loci', line):
                    multi_reads_list.append(line.strip().split()[-1])
        unique_reads = int(unique_reads_list[0])
        unique_reads_fraction = float(unique_reads_list[1].strip('%')) / 100
        multi_reads = int(multi_reads_list[0])
        multi_reads_fraction = float(multi_reads_list[1].strip('%')) / 100

        self.add_metric(
            name='Genome',
            value=self.genome['genome_name'],
        )
        self.add_metric(
            name=f'Uniquely Mapped {self.stat_prefix}',
            value=unique_reads,
            fraction=unique_reads_fraction,
        )
        self.add_metric(
            name=f'Multi-Mapped {self.stat_prefix}',
            value=multi_reads,
            fraction=multi_reads_fraction,
        )


def get_opts_star_mixin(parser, sub_program):
    parser.add_argument(
        '--genomeDir',
        help=HELP_DICT['genomeDir'],
    )
    parser.add_argument(
        '--outFilterMatchNmin',
        help="""Default `0`. Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=0
    )
    parser.add_argument(
        '--out_unmapped',
        help='Output unmapped reads.',
        action='store_true'
    )
    parser.add_argument('--STAR_param', help='Other STAR parameters.', default="")
    parser.add_argument(
        '--outFilterMultimapNmax',
        help='Default `1`. How many places are allowed to match a read at most.',
        default=1
    )
    parser.add_argument(
        '--starMem',
        help='Default `30`. Maximum memory that STAR can use.',
        default=30
    )
    if sub_program:
        parser.add_argument('--fq', help="Required. R2 fastq file.", required=True)
        parser.add_argument("--consensus_fq", action='store_true', help="A indicator that the input fastq has been consensused.")
        parser = s_common(parser)




import os
import re
import subprocess

import pysam
from celescope.rna.mkref import parse_genomeDir_rna
from celescope.tools.step import Step, s_common
from celescope.tools.utils import add_log, format_number, get_id_name_dict


class FeatureCounts(Step):
    """
    Features

    - Assigning uniquely mapped reads to genomic features with FeatureCounts.

    Output
    - `{sample}` Numbers of reads assigned to features (or meta-features).

    - `{sample}_summary` Stat info for the overall summrization results, including number of 
    successfully assigned reads and number of reads that failed to be assigned due to 
    various reasons (these reasons are included in the stat info).

    - `{sample}_Aligned.sortedByCoord.out.bam.featureCounts.bam` featureCounts output BAM, 
    sorted by coordinates；BAM file contains tags as following(Software Version>=1.1.8):
        - CB cell barcode
        - UB UMI
        - GN gene name
        - GX gene id

    - `{sample}_name_sorted.bam` featureCounts output BAM, sorted by read name.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.gtf = parse_genomeDir_rna(self.args.genomeDir)['gtf']
        self.featureCounts_param = args.featureCounts_param

        # out files
        input_basename = os.path.basename(self.args.input)
        self.featureCounts_bam = f'{self.outdir}/{input_basename}.featureCounts.bam'
        self.name_sorted_bam = f'{self.out_prefix}_name_sorted.bam'
        self.featureCount_log_file = f'{self.out_prefix}.summary'

    def format_stat(self):
        tmp_arr = []
        fh = open(self.featureCount_log_file, 'r')
        with open(self.stat_file, 'w') as stat_fh:
            p1 = re.compile(r'Assigned.*?(\d+)', flags=re.S)
            p2 = re.compile(r'Unassigned_NoFeatures.*?(\d+)', flags=re.S)
            p3 = re.compile(r'Unassigned_Ambiguity.*?(\d+)', flags=re.S)
            for line in fh:
                if line.strip() == '':
                    continue

                m1 = p1.search(line.strip())
                if m1:
                    tmp_arr.append(int(m1.group(1)))

                m2 = p2.search(line)
                if m2:
                    tmp_arr.append(int(m2.group(1)))

                m3 = p3.search(line)
                if m3:
                    tmp_arr.append(int(m3.group(1)))

            total = sum(tmp_arr)
            tmp_arr = [
                '%s(%.2f%%)' %
                (format_number(n), (n + 0.0) / total * 100) for n in tmp_arr]
            for t, s in zip(['Assigned', 'Unassigned_NoFeatures',
                            'Unassigned_Ambiguity'], tmp_arr):
                stat_fh.write('%s: %s\n' % (t, s))
        fh.close()

    @add_log
    def run_featureCounts(self):
        cmd = (
            'featureCounts '
            '-s 1 '
            f'-a {self.gtf} '
            f'-o {self.out_prefix} '  # not bam
            '-R BAM '
            f'-T {self.thread} '
            f'-t {self.args.gtf_type} '
            f'{self.args.input} '
        )
        if self.featureCounts_param:
            cmd += (" " + self.featureCounts_param)
        FeatureCounts.run_featureCounts.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @add_log
    def name_sort_bam(self):
        cmd = (
            'samtools sort -n '
            f'-@ {self.thread} '
            f'-o {self.name_sorted_bam} '
            f'{self.featureCounts_bam} '
        )
        FeatureCounts.name_sort_bam.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.run_featureCounts()
        add_tag(self.featureCounts_bam, self.gtf)
        self.name_sort_bam()
        self.format_stat()
        self.clean_up()


@add_log
def add_tag(bam, gtf):
    """
    - CB cell barcode
    - UB UMI
    - GN gene name
    - GX gene id
    """
    id_name = get_id_name_dict(gtf)
    temp_bam_file = bam + ".temp"

    with pysam.AlignmentFile(bam, "rb") as original_bam:
        header = original_bam.header
        with pysam.AlignmentFile(temp_bam_file, "wb", header=header) as temp_bam:
            for read in original_bam:
                attr = read.query_name.split('_')
                barcode = attr[0]
                umi = attr[1]
                read.set_tag(tag='CB', value=barcode, value_type='Z')
                read.set_tag(tag='UB', value=umi, value_type='Z')
                # assign to some gene
                if read.has_tag('XT'):
                    gene_id = read.get_tag('XT')
                    # if multi-mapping reads are included in original bam,
                    # there are multiple gene_ids
                    if ',' in gene_id:
                        gene_name = [id_name[i] for i in gene_id.split(',')]
                        gene_name = ','.join(gene_name)
                    else:
                        gene_name = id_name[gene_id]
                    read.set_tag(tag='GN', value=gene_name, value_type='Z')
                    read.set_tag(tag='GX', value=gene_id, value_type='Z')
                temp_bam.write(read)
    cmd = f'mv {temp_bam_file} {bam}'
    subprocess.check_call(cmd, shell=True)


@add_log
def featureCounts(args):
    step_name = "featureCounts"
    featureCounts_obj = FeatureCounts(args, step_name)
    featureCounts_obj.run()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument(
        '--gtf_type',
        help='Specify feature type in GTF annotation',
        default='exon'
    )
    parser.add_argument('--genomeDir', help='Required. Genome directory.')
    parser.add_argument('--featureCounts_param', help='Other featureCounts parameters', default="")

    if sub_program:
        parser.add_argument('--input', help='Required. BAM file path.', required=True)
        parser = s_common(parser)
    return parser



import subprocess

import pandas as pd

import celescope.tools.utils as utils
from celescope.__init__ import ROOT_PATH
from celescope.tools.star_mixin import StarMixin, get_opts_star_mixin
from celescope.tools.step import Step


class Star_rna(Step, StarMixin):
    """
    Features
    - Align R2 reads to the reference genome with STAR.
    - Collect Metrics with Picard.

    Output
    - `{sample}_Aligned.sortedByCoord.out.bam` BAM file contains Uniquely Mapped Reads.

    - `{sample}_SJ.out.tab` SJ.out.tab contains high confidence collapsed splice junctions in tab-delimited format.

    - `{sample}_Log.out` Main log with a lot of detailed information about the run. 
    This is most useful for troubleshooting and debugging.

    - `{sample}_Log.progress.out` Report job progress statistics, such as the number of processed reads, 
    % of mapped reads etc. It is updated in 1 minute intervals.

    - `{sample}_Log.Log.final.out` Summary mapping statistics after mapping job is complete, 
    very useful for quality control. The statistics are calculated for each read (single- or paired-end) and 
    then summed or averaged over all reads. Note that STAR counts a paired-end read as one read, 
    (unlike the samtools agstat/idxstats, which count each mate separately). 
    Most of the information is collected about the UNIQUE mappers 
    (unlike samtools agstat/idxstats which does not separate unique or multi-mappers). 
    Each splicing is counted in the numbers of splices, which would correspond to 
    summing the counts in SJ.out.tab. The mismatch/indel error rates are calculated on a per base basis, 
    i.e. as total number of mismatches/indels in all unique mappers divided by the total number of mapped bases.

    - `{sample}_region.log` Picard CollectRnaSeqMetrics results.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        StarMixin.__init__(self, args)
        # parse
        self.refflat = f"{self.genomeDir}/{self.genome['refflat']}"

        self.ribo_log = f'{self.out_prefix}_ribo_log.txt'
        self.ribo_run_log = f'{self.out_prefix}_ribo_run.log'
        self.picard_region_log = f'{self.out_prefix}_region.log'
        self.plot = None
        self.stats = pd.Series()

    def add_other_metrics(self):
        """
        add picard region bases
        add region plot
        if debug, add ribosomal RNA reads percent
        """

        with open(self.picard_region_log, 'r') as picard_log:
            region_dict = {}
            for line in picard_log:
                if not line:
                    break
                if line.startswith('## METRICS CLASS'):
                    header = picard_log.readline().strip().split('\t')
                    data = picard_log.readline().strip().split('\t')
                    region_dict = dict(zip(header, data))
                    break

        total = float(region_dict['PF_ALIGNED_BASES'])
        exonic_regions = int(region_dict['UTR_BASES']) + \
            int(region_dict['CODING_BASES'])
        intronic_regions = int(region_dict['INTRONIC_BASES'])
        intergenic_regions = int(region_dict['INTERGENIC_BASES'])

        self.add_metric(
            name='Base Pairs Mapped to Exonic Regions',
            value=exonic_regions,
            total=total,
        )
        self.add_metric(
            name='Base Pairs Mapped to Intronic Regions',
            value=intronic_regions,
            total=total,
        )
        self.add_metric(
            name='Base Pairs Mapped to Intergenic Regions',
            value=intergenic_regions,
            total=total,
        )

        # ribo
        if self.debug:
            with open(self.ribo_log, 'r') as ribo_log:
                for line in ribo_log:
                    if line.find('#Matched') != -1:
                        items = line.split()
                        Reads_Mapped_to_rRNA = int(items[1])
                    if line.find('#Total') != -1:
                        items = line.split()
                        Reads_Total = int(items[1])
                self.add_metric(
                    name=f'{self.stat_prefix} Mapped to rRNA',
                    value=Reads_Mapped_to_rRNA,
                    total=Reads_Total,
                )

        region_plot = {'region_labels': ['Exonic Regions', 'Intronic Regions', 'Intergenic Regions'],
                       'region_values': [exonic_regions, intronic_regions, intergenic_regions]}
        self.add_content_item("data", STAR_plot=region_plot)

    @utils.add_log
    def ribo(self):
        # TODO remove bbduk.sh and use picard ribo bases
        human_ribo_fa = f'{ROOT_PATH}/data/rRNA/human_ribo.fasta'
        cmd = (
            f'bbduk.sh '
            f'in1={self.fq} '
            f'ref={human_ribo_fa} '
            f'stats={self.ribo_log} '
            f'overwrite=t '
            f'> {self.ribo_run_log} 2>&1 '
        )
        Star_rna.ribo.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def picard(self):
        cmd = [
            'picard',
            '-Xmx20G',
            '-XX:ParallelGCThreads=4',
            'CollectRnaSeqMetrics',
            'I=%s' % (self.STAR_bam),
            'O=%s' % (self.picard_region_log),
            'REF_FLAT=%s' % (self.refflat),
            'STRAND=NONE',
            'VALIDATION_STRINGENCY=SILENT']
        cmd_str = ' '.join(cmd)
        Star_rna.picard.logger.info(cmd_str)
        subprocess.check_call(cmd)

    @utils.add_log
    def run(self):
        self.run_star()
        self.picard()
        if self.debug:
            self.ribo()
        self.add_other_metrics()
        self.clean_up()


def star(args):
    step_name = "star"
    runner = Star_rna(args, step_name)
    runner.run()


def get_opts_star(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)



import pandas as pd

import celescope.tools.utils as utils
from celescope.tools.step import Step
from celescope.tools.analysis_mixin import AnalysisMixin
from celescope.rna.analysis import get_opts_analysis


class Analysis_rna_virus(Step, AnalysisMixin):
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        AnalysisMixin.__init__(self, args)

        # set
        self.virus_df = pd.read_csv(args.virus_file, sep="\t")

        self.auto_assign_bool = False
        if args.type_marker_tsv and args.type_marker_tsv != 'None':
            self.auto_assign_bool = True
            self.save_rds = True
    
    
    def run(self):
        self.seurat(self.args.matrix_file, self.args.save_rds, self.args.genomeDir)
        if self.auto_assign_bool:
            self.auto_assign(self.type_marker_tsv)
        self.run_analysis()
        virus_tsne = self.virus_tsne_list()
        self.add_data_item(cluster_tsne=self.cluster_tsne)
        self.add_data_item(virus_tsne=virus_tsne)
        self.add_data_item(table_dict=self.table_dict)

        self.clean_up()


    def virus_tsne_list(self):
        """
        return data dic
        """
        df = pd.merge(self.tsne_df, self.virus_df, on="barcode", how="left")
        df["UMI"] = df["UMI"].fillna(0)
        tSNE_1 = list(df.tSNE_1)
        tSNE_2 = list(df.tSNE_2)
        virus_UMI = list(df.UMI)
        res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "virus_UMI": virus_UMI}
        return res


@utils.add_log
def analysis_rna_virus(args):

    step_name = "analysis_rna_virus"
    runner = Analysis_rna_virus(args, step_name)
    runner.run()


def get_opts_analysis_rna_virus(parser, sub_program):
    get_opts_analysis(parser, sub_program)
    if sub_program:
        parser.add_argument(
            '--virus_file',
            help='virus UMI count file',
            required=True)



import configparser
import subprocess

import pandas as pd
import pysam
from venn import generate_petal_labels,draw_venn,generate_colors

import celescope.tools.utils as utils
from celescope.tools.analysis_mixin import AnalysisMixin
from celescope.tools.step import Step, s_common
from celescope.snp.variant_calling import read_CID
from celescope.__init__ import HELP_DICT


class Analysis_variant(Step, AnalysisMixin):
    """
    Features
    - Annotate variants with [Annovar](https://annovar.openbioinformatics.org/en/latest/).

    Output
    `{sample}.{genome_version}_multianno.txt` Annovar main output file. `CID` and `VID` are added to the `Otherinfo` column.

    `{sample}_variant_table.tsv` Formatted `multianno` file with `ncell_cover`and `ncell_alt` added.
    
    `{sample}_variant_top5.jpg` The Venn diagram of the 5 variants with the highest `ncell_alt`.

    `{sample}_variant_ncell.tsv` Number of cells with read count at each variant's position. 
    - `VID`: Variant ID. 
    - `ncell_cover`: number of cells with read count at this position. 
    - `ncell_alt`: number of cells with variant read count only. 
    - `ncell_ref`: number of cells with reference read count only. 
    - `ncell_ref_and_alt`: number of cells with both variant and reference read count.
    - `RID`: Target region ID. This column will be added when `--panel` option were provided.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        AnalysisMixin.__init__(self, args)
        self.filter_variant_count_file = args.filter_variant_count_file
        self.CID_file = args.CID_file
        self.filter_vcf = args.filter_vcf
        self.annovar_config = args.annovar_config
        self.match_dir = args.match_dir

        # parse
        self.annovar_section = self.read_annovar_config()

        # out
        self.ncell_file = f'{self.out_prefix}_variant_ncell.tsv'
        self.vcf_GT = f'{self.out_prefix}_addGT.vcf'
        buildver = self.annovar_section['buildver']
        self.annovar_file = f'{self.out_prefix}.{buildver}_multianno.txt'

    @utils.add_log
    def ncell_metrics(self):
        variant_count_df = pd.read_table(self.filter_variant_count_file, sep='\t')
        variant_count_df = variant_count_df.loc[(variant_count_df["ref_count"]!=0) | 
            (variant_count_df["alt_count"]!=0)]
        if 'RID' in variant_count_df.columns: 
            df_cover = variant_count_df.groupby('VID').agg({'RID':'first','CID':'count'})
        else:
            df_cover = variant_count_df.groupby('VID').agg({'CID': 'count'})
        df_cover.rename(columns={'CID': 'ncell_cover'}, inplace=True)

        df_ref = variant_count_df.loc[(variant_count_df["ref_count"]!=0) & 
            (variant_count_df["alt_count"]==0)].groupby('VID').agg({'CID':'count'})
        df_ref.rename(columns={'CID': 'ncell_ref'}, inplace=True)

        df_alt = variant_count_df.loc[(variant_count_df["ref_count"]==0) & 
            (variant_count_df["alt_count"]!=0)].groupby('VID').agg({'CID':'count'})
        df_alt.rename(columns={'CID': 'ncell_alt'}, inplace=True)

        df_ref_and_alt = variant_count_df.loc[(variant_count_df["ref_count"]!=0) & 
            (variant_count_df["alt_count"]!=0)].groupby('VID').agg({'CID':'count'})
        df_ref_and_alt.rename(columns={'CID': 'ncell_ref_and_alt'}, inplace=True)

        df_list = [df_ref, df_alt, df_ref_and_alt]
        df_ncell = df_cover
        for df in df_list:
            df_ncell = pd.merge(df_ncell, df, on='VID', how='left')
        df_ncell.fillna(0, inplace=True)

        # dtype after fillna is float
        float_cols = ['ncell_ref', 'ncell_alt', 'ncell_ref_and_alt']
        for col in float_cols:
            df_ncell[col] = df_ncell[col].astype('int')

        df_ncell.to_csv(self.ncell_file, sep = '\t',index = True)

    def get_df_count_tsne(self):
        '''
        output: f'{self.outdir}/{self.sample}_count_tsne.tsv'
        '''
        df_vc = pd.read_csv(self.filter_variant_count_file, sep='\t')
        df_vc = df_vc[df_vc["alt_count"] > 0]
        df_vc_cell = df_vc.groupby('CID').agg({
            'alt_count': 'count',
            'VID': list,
        })

        df_CID, _df_valid = read_CID(self.CID_file)
        df_CID = df_CID.reset_index()
        tsne_df_CID = pd.merge(self.tsne_df, df_CID, on='barcode', how='left')

        df_vc_barcode = pd.merge(df_vc_cell, df_CID, on='CID')
        df_vc_barcode_tsne = pd.merge(df_vc_barcode, tsne_df_CID, on=['barcode', 'CID'], how='right')
        df_vc_barcode_tsne['value'] = df_vc_barcode_tsne['alt_count']
        df_vc_barcode_tsne['value'] = df_vc_barcode_tsne['value'].fillna(0)
        df_vc_barcode_tsne['value'].astype('int32')
        df_count_tsne = df_vc_barcode_tsne
        # out
        out_file = f'{self.outdir}/{self.sample}_count_tsne.tsv'
        df_out = df_vc_barcode_tsne[df_vc_barcode_tsne['value'] > 0]
        cols = ['CID', 'alt_count', 'VID', 'barcode', 'cluster', 'tSNE_1', 'tSNE_2', 'Gene_Counts']
        df_out = df_out[cols]
        df_out.to_csv(out_file, sep='\t')
        return df_count_tsne

    def get_count_tsne(self, df_count_tsne):
        def return_text(row):
            text = f'CID:{row["CID"]} <br>Variants:{str(int(row["value"]))} <br>VID:{row["VID"]}'
            return text
        tSNE_1 = list(df_count_tsne.tSNE_1)
        tSNE_2 = list(df_count_tsne.tSNE_2)
        text = list(df_count_tsne.apply(return_text, axis=1))
        value = list(df_count_tsne.value)
        title = 't-SNE plot Colored by Cell Variant Counts'
        count_tsne = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "text": text, 'value': value, 'title': title}
        return count_tsne

    def add_GT(self):
        '''
        add genotype to VCF file to avoid vcf parse error
        '''
        with pysam.VariantFile(self.filter_vcf, 'r') as vcf:
            with pysam.VariantFile(self.vcf_GT, 'w', header=vcf.header) as vcf_GT:
                for rec in vcf:
                    for sample in rec.samples:
                        rec.samples[sample]["GT"] = (1, 1)
                        vcf_GT.write(rec)

    def get_df_table(self):

        df_vcf = utils.parse_vcf(self.vcf_GT, infos=['VID', 'CID'])
        df_annovar = self.annovar()
        df_vcf = pd.concat((df_vcf, df_annovar), axis=1)
        ncell_df = pd.read_table(self.ncell_file,sep = "\t")
        ncell_df.loc[:,"VID"] = ncell_df.loc[:,"VID"].astype(str)
        df_vcf["nCell"] = df_vcf["CID"].apply(func=lambda row: 1 if isinstance(row, str) else len(row))

        df_vcf = pd.merge(left = df_vcf,
                          right = ncell_df,
                          on = "VID",
                          how = "left")

        out_df_vcf = f'{self.outdir}/{self.sample}_variant_table.tsv'
        df_vcf.drop("nCell",axis = 1).to_csv(out_df_vcf, sep='\t', index=False)

        cols = ['VID', "CID",'Chrom', 'Pos', 'Alleles', 'Gene',  
            'ncell_cover', "ncell_ref", 'ncell_alt', "ncell_ref_and_alt",
            'mRNA', 'Protein', 'COSMIC']
        df_vcf = df_vcf[cols]
        return df_vcf
    
    def get_venn_plot(self):
        df_top_5 = self.get_df_table().sort_values(by = "ncell_alt",ascending=False).iloc[:5,:]
        plot = {}
        cid_lst = df_top_5.loc[:,"CID"].to_list()
        vid_lst = df_top_5.loc[:,"VID"].to_list()
        for cid,vid in zip(cid_lst,vid_lst):
            plot[f"VID_{vid}"] = set(cid)
        share_cid  = list(set.intersection(*map(set,cid_lst)))
        if share_cid == []:
            share_cid.append("None")
        #venn plot
        set_cid = list(plot.values())
        set_name = list(plot.keys())
        labels = generate_petal_labels(set_cid)
        plot = draw_venn(
                         petal_labels=labels, 
                         dataset_labels=set_name,
                         hint_hidden=False,
                         colors=generate_colors(n_colors=5), 
                         figsize=(8, 8),
                         fontsize=14, 
                         legend_loc="best",
                         ax=None
                         )
        fig = plot.get_figure()
        fig.savefig(f'{self.outdir}/{self.sample}_variant_top5.jpg',dpi = 600)
        pd.DataFrame({"top5_variant_shared_cells":share_cid}).to_csv(f'{self.outdir}/{self.sample}_top5_shared_cells.tsv',sep = '\t',index = None)

    def run(self):
        self.add_GT()
        self.ncell_metrics()
        cluster_tsne = self.get_cluster_tsne(colname='cluster', tsne_df=self.tsne_df)
        df_count_tsne = self.get_df_count_tsne()
        count_tsne = self.get_count_tsne(df_count_tsne)
        df_vcf = self.get_df_table()
        table_dict = Step.get_table(title='Variant table', table_id='variant_table', df_table=df_vcf.drop(["CID"],axis = 1))

        self.add_data_item(cluster_tsne=cluster_tsne)
        self.add_data_item(count_tsne=count_tsne)
        self.add_data_item(table_dict=table_dict)
        self.clean_up()
        self.get_venn_plot()

    def read_annovar_config(self):
        '''
        read annovar config file
        '''
        config = configparser.ConfigParser()
        config.read(self.annovar_config)
        section = config['ANNOVAR']
        return section

    @utils.add_log
    def annovar(self):

        section = self.annovar_section
        annovar_dir = section['dir']
        db = section['db']
        buildver = section['buildver']
        protocol = section['protocol']
        operation = section['operation']

        # convert
        input_file = f'{self.outdir}/{self.sample}.input'
        cmd = (
            f'perl {annovar_dir}/convert2annovar.pl '
            f'-format vcf4 '
            f'--includeinfo '
            f'{self.vcf_GT} > {input_file}'
        )
        subprocess.check_call(cmd, shell=True)

        # annotate
        cmd = (
            f'perl {annovar_dir}/table_annovar.pl '
            f'{input_file} '
            f'{db} '
            f'-buildver {buildver} '
            f'-protocol {protocol} '
            f'-operation {operation} '
            f'-out {self.outdir}/{self.sample} '
            f'--otherinfo '
        )
        Analysis_variant.annovar.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

        # df
        df_annovar = utils.parse_annovar(self.annovar_file)
        return df_annovar


@utils.add_log
def analysis_snp(args):
    step = 'analysis_snp'
    step_snp = Analysis_variant(args, step)
    step_snp.run()


def get_opts_analysis_snp(parser, sub_program):
    parser.add_argument('--annovar_config', help='ANNOVAR config file.', required=True)
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser.add_argument('--filter_vcf', help='filter vcf file.', required=True)
        parser.add_argument('--CID_file', help='CID_file.', required=True)
        parser.add_argument('--filter_variant_count_file', help='filter variant count file.', required=True)




import celescope.tools.utils as utils
from celescope.tools.star_mixin import StarMixin, get_opts_star_mixin
from celescope.tools.step import Step


class StarFusion(Step, StarMixin):
    def __init__(self, args, step_name):
        args.genomeDir = args.fusion_genomeDir
        Step.__init__(self, args, step_name)
        StarMixin.__init__(self, args)

    def run(self):
        self.run_star()
        self.clean_up()


@utils.add_log
def star_fusion(args):
    step_name = "star_fusion"
    runner = StarFusion(args, step_name)
    runner.run()


def get_opts_star_fusion(parser, sub_program):
    get_opts_star_mixin(parser, sub_program)
    # will cause `conflicting option string: --genomeDir`
    # parser.add_argument('--genomeDir', help=argparse.SUPPRESS)
    parser.add_argument('--fusion_genomeDir', help='fusion gene STAR index genome directory', required=True)



from celescope.tcr_fl.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_tcr_fl(Multi):

    def split_fq(self, sample):
        step = 'split_fq'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd = (
            f'{self.__APP__} '
            f'{self.__ASSAY__} '
            f'{step} '
            f'--outdir {self.outdir_dic[sample][step]} '
            f'--sample {sample} '
            f'--assay {self.__ASSAY__} '
            f'--fq {fq} '
            f'--nCell {self.nCell} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def assemble(self, sample):
        step = 'assemble'
        fastq_dir = f'{self.outdir_dic[sample]["split_fq"]}/fastq'
        cmd = (
            f'{self.__APP__} '
            f'{self.__ASSAY__} '
            f'{step} '
            f'--outdir {self.outdir_dic[sample][step]} '
            f'--sample {sample} '
            f'--assay {self.__ASSAY__} '
            f'--fastq_dir {fastq_dir} '
            f'--thread {self.thread} '
        )
        self.process_cmd(cmd, step, sample, m=4 * int(self.args.thread), x=self.args.thread)


def main():
    multi = Multi_tcr_fl(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



from celescope.rna_virus.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_rna_virus(Multi):

    def star_virus(self, sample):
        step = 'star_virus'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '

        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_virus(self, sample):
        step = 'count_virus'
        barcode_file = f'{self.outdir_dic[sample]["count"]}/{sample}_matrix_10X/barcodes.tsv'
        virus_bam = f'{self.outdir_dic[sample]["star_virus"]}/{sample}_virus_Aligned.sortedByCoord.out.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--virus_bam {virus_bam} '
            f'--barcode_file {barcode_file} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def analysis_rna_virus(self, sample):
        step = 'analysis_rna_virus'
        virus_file = f'{self.outdir_dic[sample]["count_virus"]}/{sample}_virus_UMI_count.tsv'
        matrix_file = f'{self.outdir_dic[sample]["count"]}/{sample}_matrix_10X'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--virus_file {virus_file} '
            f'--matrix_file {matrix_file} '

        )
        self.process_cmd(cmd, step, sample, m=15, x=1)


def main():
    multi = Multi_rna_virus(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import subprocess
import unittest
from collections import defaultdict
from itertools import groupby

import numpy as np
import pysam
from xopen import xopen

import celescope.tools.utils as utils
from celescope.tools.step import Step, s_common


class Consensus(Step):
    """
    Features
    - Consensus all the reads of the same (barcode, UMI) combinations into one read(UMI).

    Output
    - `{sample}_consensus.fq` Consensus fastq.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.min_consensus_read = int(self.args.min_consensus_read)

        # out files
        self.fq_tmp_file = f'{self.out_prefix}_sorted.fq.tmp'
        self.consensus_fq = f'{self.out_prefix}_consensus.fq'

    @utils.add_log
    def run(self):
        if self.args.not_consensus:
            Consensus.run.logger.warning("Will not perform UMI consensus!")
            return

        sort_fastq(self.args.fq, self.fq_tmp_file, self.outdir)
        n, total_ambiguous_base_n, length_list = sorted_dumb_consensus(
            fq=self.fq_tmp_file,
            outfile=self.consensus_fq,
            threshold=self.args.threshold,
            min_consensus_read=self.min_consensus_read,
        )

        self.add_metric(
            name="UMI Counts",
            value=n,
        )
        self.add_metric(
            name="Mean UMI Length",
            value=round(np.mean(length_list), 2),
        )
        self.add_metric(
            name="Ambiguous Base Counts",
            value=total_ambiguous_base_n,
            total=sum(length_list),
        )
        self.clean_up()


@utils.add_log
def sort_fastq(fq, fq_tmp_file, outdir):
    tmp_dir = f'{outdir}/tmp'
    cmd = (
        f'mkdir {tmp_dir};'
        f'less {fq} | paste - - - - | sort -T {tmp_dir} -k1,1 -t " " | tr "\t" "\n" > {fq_tmp_file};'
    )
    subprocess.check_call(cmd, shell=True)


@utils.add_log
def sorted_dumb_consensus(fq, outfile, threshold, min_consensus_read):
    '''
    consensus read in name-sorted fastq
    output (barcode,umi) consensus fastq
    '''
    read_list = []
    n_umi = 0
    total_ambiguous_base_n = 0
    length_list = []
    out_h = xopen(outfile, 'w')

    def keyfunc(read):
        attr = read.name.split('_')
        return (attr[0], attr[1])

    with pysam.FastxFile(fq) as fh:
        for (barcode, umi), g in groupby(fh, key=keyfunc):
            read_list = []
            for read in g:
                read_list.append([read.sequence, read.quality])
            consensus_seq, consensus_qual, ambiguous_base_n, con_len = dumb_consensus(
                read_list,
                threshold=threshold,
                min_consensus_read=min_consensus_read,
                ambiguous="N"
            )
            n_umi += 1
            prefix = "_".join([barcode, umi])
            read_name = f'{prefix}_{n_umi}'
            out_h.write(utils.fastq_line(read_name, consensus_seq, consensus_qual))
            if n_umi % 10000 == 0:
                sorted_dumb_consensus.logger.info(f'{n_umi} UMI done.')
            total_ambiguous_base_n += ambiguous_base_n
            length_list.append(con_len)

    out_h.close()
    return n_umi, total_ambiguous_base_n, length_list


def dumb_consensus(read_list, threshold=0.5, min_consensus_read=1, ambiguous='N', default_qual='F'):
    '''
    This is similar to biopython dumb_consensus.
    It will just go through the sequence residue by residue and count up the number of each type
    of residue (ie. A or G or T or C for DNA) in all sequences in the
    alignment. If 
    1. the percentage of the most common residue type > threshold;
    2. most common residue reads >= min_consensus_read;
    then we will add that residue type,
    otherwise an ambiguous character will be added.
    elements of read_list: [entry.sequence,entry.quality]
    '''

    con_len = get_read_length(read_list, threshold=threshold)
    consensus_seq = ""
    consensus_qual = ""
    ambiguous_base_n = 0
    for n in range(con_len):
        atom_dict = defaultdict(int)
        quality_dict = defaultdict(int)
        num_atoms = 0
        for read in read_list:
            # make sure we haven't run past the end of any sequences
            # if they are of different lengths
            sequence = read[0]
            quality = read[1]
            if n < len(sequence):
                atom = sequence[n]
                atom_dict[atom] += 1
                num_atoms = num_atoms + 1

                base_qual = quality[n]
                quality_dict[base_qual] += 1

        consensus_atom = ambiguous
        for atom in atom_dict:
            if atom_dict[atom] > num_atoms * threshold and atom_dict[atom] >= min_consensus_read:
                consensus_atom = atom
                break
        if consensus_atom == ambiguous:
            ambiguous_base_n += 1
        consensus_seq += consensus_atom

        max_freq_qual = 0
        consensus_base_qual = default_qual
        for base_qual in quality_dict:
            if quality_dict[base_qual] > max_freq_qual:
                max_freq_qual = quality_dict[base_qual]
                consensus_base_qual = base_qual

        consensus_qual += consensus_base_qual
    return consensus_seq, consensus_qual, ambiguous_base_n, con_len


def get_read_length(read_list, threshold=0.5):
    '''
    compute read_length from read_list. 
    length = max length with read fraction >= threshold
    elements of read_list: [entry.sequence,entry.quality]
    '''

    n_read = len(read_list)
    length_dict = defaultdict(int)
    for read in read_list:
        length = len(read[0])
        length_dict[length] += 1
    for length in length_dict:
        length_dict[length] = length_dict[length] / n_read

    fraction = 0
    for length in sorted(length_dict.keys(), reverse=True):
        fraction += length_dict[length]
        if fraction >= threshold:
            return length


@utils.add_log
def consensus(args):

    step_name = "consensus"
    consensus_obj = Consensus(args, step_name)
    consensus_obj.run()


def get_opts_consensus(parser, sub_program):
    parser.add_argument("--threshold", help='Default 0.5. Valid base threshold. ', type=float, default=0.5)
    parser.add_argument("--not_consensus", help="Skip the consensus step. ", action='store_true')
    parser.add_argument("--min_consensus_read", help="Minimum number of reads to support a base. ", default=1)
    if sub_program:
        parser.add_argument("--fq", help="Required. Fastq file.", required=True)
        s_common(parser)


class Consensus_unittest(unittest.TestCase):

    def test_get_read_length(self):
        read_list = [['AAAA', 'FFFF'], ['TTT', 'FFF'], ['CCC', 'FFF'], ['GGGGGGG', 'FFFFFFF']]
        assert get_read_length(read_list, 0.5) == 4

    def test_dumb_consensus(self):
        read_list = [('AAAA', 'FFFF'), ('TTT', 'FF;'), ('CCCC', 'FFFF'), ('GGGAGGG', 'FFFFFFF')]
        consensus_seq, _consensus_qual, _ambiguous_base_n, _con_len = dumb_consensus(read_list, 0.5)
        self.assertEqual(consensus_seq, 'NNNA')

        read_list = [('AAAA', 'FFFF'), ('TTT', 'FF;'), ('CCC', 'FFF'), ('GGGGGGG', 'FFFFFFF')]
        consensus_seq, _consensus_qual, _ambiguous_base_n, _con_len = dumb_consensus(read_list, 0.5)
        self.assertEqual(consensus_seq, 'NNNN')


if __name__ == '__main__':
    unittest.main()


import subprocess
from collections import defaultdict
from multiprocessing import Pool
from itertools import groupby
from functools import partial
from concurrent.futures import ProcessPoolExecutor

import pandas as pd
import numpy as np
import pyranges as pr
import pysam
from scipy.io import mmwrite
from scipy.sparse import coo_matrix

import celescope.tools.utils as utils
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.rna.mkref import parse_genomeDir_rna


OTSU_READ_MIN = 30


def parse_vcf(vcf_file, cols=('chrom', 'pos', 'alleles',), infos=('VID',)):
    '''
    parse vcf into df
    '''
    vcf = pysam.VariantFile(vcf_file)
    rec_dict_list = []
    for rec in vcf.fetch():
        rec_dict = {}
        for col in cols:
            rec_dict[col] = getattr(rec, col)
            # if ref == alt: alleles=(ref,)
            # else alleles=(ref, alt)
            if col == 'alleles':
                rec_dict['ref'] = rec_dict['alleles'][0]
                rec_dict['alt'] = '.'
                if len(rec_dict['alleles']) >= 2:
                    rec_dict['alt'] = ','.join(rec_dict['alleles'][1:])

        for info in infos:
            rec_dict[info] = rec.info[info]
        rec_dict_list.append(rec_dict)

    df = pd.DataFrame.from_dict(rec_dict_list)
    vcf.close()
    return df


def read_CID(CID_file):
    df_index = pd.read_csv(CID_file, sep='\t', index_col=0).reset_index()
    df_valid = df_index[df_index['valid'] == True]
    return df_index, df_valid


@utils.add_log
def call_snp(outdir, fasta, CID):

    call_snp.logger.info('Processing Cell %s' % CID)
    bam = f'{outdir}/cells/cell{CID}/cell{CID}.bam'
    # sort
    sorted_bam = f'{outdir}/cells/cell{CID}/cell{CID}_sorted.bam'
    cmd_sort = (
        f'samtools sort {bam} -o {sorted_bam}'
    )
    subprocess.check_call(cmd_sort, shell=True)

    # mpileup
    bcf = f'{outdir}/cells/cell{CID}/cell{CID}.bcf'
    cmd_mpileup = (
        f'bcftools mpileup -Ou '
        f'-f {fasta} '
        f'{sorted_bam} -o {bcf} '
    )
    subprocess.check_call(cmd_mpileup, shell=True)

    # call
    out_vcf = f'{outdir}/cells/cell{CID}/cell{CID}.vcf'
    cmd_call = (
        f'bcftools call -mv -Ov '
        f'-o {out_vcf} '
        f'{bcf}'
        f'>/dev/null 2>&1 '
    )
    subprocess.check_call(cmd_call, shell=True)

    # norm
    norm_vcf = f'{outdir}/cells/cell{CID}/cell{CID}_norm.vcf'
    cmd_norm = (
        f'bcftools norm -d none '
        f'-f {fasta} '
        f'{out_vcf} '
        f'-o {norm_vcf} '
    )
    subprocess.check_call(cmd_norm, shell=True)

    # call all position
    out_all_vcf = f'{outdir}/cells/cell{CID}/cell{CID}_all.vcf'
    cmd_all_call = (
        f'bcftools call -m -Ov '
        f'-o {out_all_vcf} '
        f'{bcf}'
        f'>/dev/null 2>&1 '
    )
    subprocess.check_call(cmd_all_call, shell=True)

    # norm all
    norm_all_vcf = f'{outdir}/cells/cell{CID}/cell{CID}_all_norm.vcf'
    cmd_all_norm = (
        f'bcftools norm -d none '
        f'-f {fasta} '
        f'{out_all_vcf} '
        f'-o {norm_all_vcf} '
    )
    subprocess.check_call(cmd_all_norm, shell=True)


def map_vcf_row(row, df_cell_vcf):
    """
    get ref and UMI for each variant
    Args:
        row: each row from merged_vcf
    """
    pos = row['pos']
    chrom = row['chrom']
    alt = row['alt']
    df_pos = df_cell_vcf[(df_cell_vcf['pos'] == pos) & (df_cell_vcf['chrom'] == chrom)]
    df_ref = df_pos[df_pos['alt'] == '.']
    df_alt = df_pos[df_pos['alt'] == alt]
    ref_UMI = 0
    alt_UMI = 0
    if df_ref.shape[0] != 0:
        ref_UMI = get_DP4(df_ref, 'ref')
    if df_alt.shape[0] != 0:
        alt_UMI = get_DP4(df_alt, 'alt')
    return ref_UMI, alt_UMI


def get_DP4(row, alt):
    DP4 = row['DP4'].iloc[0]
    if alt == 'ref':
        indexs = [0, 1]
    elif alt == 'alt':
        indexs = [2, 3]
    umi = sum([DP4[index] for index in indexs])
    return umi

@utils.add_log
def cell_UMI(CID, outdir, df_vcf):
    """
    help function to get ref and alt UMI for each (cell, variant)
    """
    cell_UMI.logger.info(str(CID) + ' cell')
    norm_all_vcf = f'{outdir}/cells/cell{CID}/cell{CID}_all_norm.vcf'
    df_cell_vcf = parse_vcf(norm_all_vcf, infos=['DP4'])

    dict_list = []
    for _index, row in df_vcf.iterrows():
        ref_UMI, alt_UMI = map_vcf_row(row, df_cell_vcf)
        if (ref_UMI + alt_UMI) != 0:
            VID = row['VID']
            dic = {
                'VID': VID,
                'CID': CID,
                'ref_count': ref_UMI,
                'alt_count': alt_UMI,
            }
            dict_list.append(dic)
    df_UMI = pd.DataFrame(dict_list)
    return df_UMI


class Variant_calling(Step):
    """
    Features
    - Perform variant calling at single cell level.

    Output

    `{sample}_VID.tsv` A unique numeric ID is assigned for each variant, 

    - `RID`: Target region ID. This column will be added when `--panel` option were provided.

    `{sample}_CID.tsv` A unique numeric ID is assigned for each cell.

    `{sample}_RID.tsv` A unique numeric ID is assigned for each target region. This file will be created when `--panel` option were provided.


    `{sample}_merged.vcf ` VCF file containing all variants of all cells. `VID` and `CID` are added to the `INFO` column.

    `{sample}_filter.vcf` VCF file after filtering. Invalid `CID`s are removed from the `INFO` column.

    `{sample}_variant_count.tsv`  Reference and variant supporting reads/UMIs counts.

    `{sample}_filter_variant_count.tsv`  Reference and variant supporting reads/UMIs counts after filtering.

    `{sample}_support.mtx` Support matrix in [Matrix Market Exchange Formats](https://math.nist.gov/MatrixMarket/formats.html). Rows 
    are variants(VID) and columns are cells(CID). The value can be 1, 2 or 3.
    
    1 : all reads/UMIs at the position support the ref allele.  
    2 : all reads/UMIs at the position support the alt allele.  
    3 : one or more reads/UMIs support both the alt and the ref allele.  
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.barcodes, _num = utils.read_barcode_file(args.match_dir)
        self.fasta = parse_genomeDir_rna(args.genomeDir)['fasta']
        self.df_vcf = None
        self.panel = args.panel
        self.panel_bool = self.panel and self.panel != "None"

        # out
        self.splitN_bam = f'{self.out_prefix}_splitN.bam'
        self.splitN_bam_name_sorted = f'{self.out_prefix}_splitN_name_sorted.bam'
        self.CID_file = f'{self.out_prefix}_CID.tsv'
        self.VID_file = f'{self.out_prefix}_VID.tsv'
        self.RID_file = f'{self.out_prefix}_RID.tsv'
        self.cells_dir = f'{self.outdir}/cells/'
        self.merged_vcf_file = f'{self.out_prefix}_merged.vcf'
        self.filter_vcf_file = f'{self.out_prefix}_filter.vcf'
        self.variant_count_file = f'{self.out_prefix}_variant_count.tsv'
        self.otsu_dir = f'{self.out_prefix}_otsu/'
        self.otsu_threshold_file = f'{self.otsu_dir}/{self.sample}_otsu_threshold.tsv'
        self.filter_variant_count_file = f'{self.out_prefix}_filter_variant_count.tsv'
        if args.min_support_read == 'auto':
            utils.check_mkdir(self.otsu_dir)
        self.support_matrix_file = f'{self.out_prefix}_support.mtx'

        # run
        self.CID_dict = defaultdict(dict)
        for index, barcode in enumerate(self.barcodes):
            CID = index + 1
            self.CID_dict[barcode]['CID'] = CID
            self.CID_dict[barcode]['valid'] = False


    @utils.add_log
    def SplitNCigarReads(self):
        cmd = (
            f'gatk '
            f'SplitNCigarReads '
            f'--do-not-fix-overhangs '
            f'-R {self.fasta} '
            f'-I {self.args.bam} '
            f'-O {self.splitN_bam} '
        )
        Variant_calling.SplitNCigarReads.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def split_bam(self):
        '''
        input:
            bam: bam from splitN
            barcodes: cell barcodes, list
        ouput:
            bam_dict: assign reads to cell barcodes and UMI
            count_dict: UMI counts per cell
            CID: assign ID(1-based) to cells
        '''

        # name sort
        samtools_runner = utils.Samtools(self.splitN_bam, self.splitN_bam_name_sorted, threads=self.thread)
        samtools_runner.sort_bam(by='name', print_log=True)

        def keyfunc(x):
            # return barcode
            return x.query_name.split('_', 1)[0]

        # read bam and split
        with pysam.AlignmentFile(self.splitN_bam_name_sorted, "rb") as samfile:
            header = samfile.header
            for barcode, g in groupby(samfile, keyfunc):
                if barcode in self.CID_dict:
                    CID = self.CID_dict[barcode]['CID']
                    self.CID_dict[barcode]['valid'] = True
                    cell_dir = f'{self.cells_dir}/cell{CID}'
                    utils.check_mkdir(cell_dir)
                    cell_bam = f'{self.cells_dir}/cell{CID}/cell{CID}.bam'
                    with pysam.AlignmentFile(cell_bam, "wb", header=header) as out_bam:
                        for read in g:
                            out_bam.write(read)

    def write_CID_file(self):
        # out CID file
        df_CID = pd.DataFrame(self.CID_dict).T
        df_CID.index.name = 'barcode'
        df_CID.to_csv(self.CID_file, sep='\t')

    @utils.add_log
    def call_all_snp(self):
        _df_index, df_valid = self.read_CID()
        CIDs = df_valid['CID']
        func = partial(call_snp, self.outdir, self.fasta)
        with Pool(self.thread) as pool:
            pool.map(func, CIDs)

    def read_CID(self):
        return read_CID(self.CID_file)

    @utils.add_log
    def merge_vcf(self):
        '''
        merge cell vcf into one non-duplicated vcf
        add VID(variant ID) and CID(cell ID)
        '''
        _df_index, df_valid = self.read_CID()
        CIDs = list(df_valid['CID'])
        # variant dict
        v_cols = ['chrom', 'pos', 'alleles']
        v_dict = dict()

        for CID in CIDs:
            CID = str(CID)
            vcf_file = f'{self.outdir}/cells/cell{CID}/cell{CID}_norm.vcf'
            with pysam.VariantFile(vcf_file, 'r') as vcf:
                for rec in vcf.fetch():
                    v = ','.join([str(getattr(rec, col)) for col in v_cols])
                    if not v in v_dict:
                        v_dict[v] = dict()
                        for col in v_cols:
                            v_dict[v][col] = getattr(rec, col)
                        v_dict[v]['CID'] = [CID]
                        v_dict[v]['record'] = rec
                    else:
                        v_dict[v]['CID'].append(CID)

        # output
        def get_vcf_header(CIDs):
            CID = CIDs[0]
            vcf_file = f'{self.outdir}/cells/cell{CID}/cell{CID}_norm.vcf'
            with pysam.VariantFile(vcf_file, 'r') as vcf:
                return vcf.header

        vcf_header = get_vcf_header(CIDs)
        vcf_header.info.add('VID', number=1, type='String', description='Variant ID')
        vcf_header.info.add('CID', number=1, type='String', description='Cell ID')
        with pysam.VariantFile(self.merged_vcf_file, 'w', header=vcf_header) as merged_vcf:
            VID = 0
            for _keys, values in sorted(v_dict.items(), 
                key=lambda x: (x[1]['chrom'], x[1]['pos'], x[1]['alleles'])):
                VID += 1
                rec = values['record']
                CID = ','.join(values['CID'])
                record = merged_vcf.new_record()
                cols = ['chrom', 'pos', 'alleles']
                for col in cols:
                    setattr(record, col, getattr(rec, col))
                record.info['VID'] = str(VID)
                record.info['CID'] = CID
                merged_vcf.write(record)

    @utils.add_log
    def write_VID_file(self):
        df_vcf = parse_vcf(self.merged_vcf_file)
        df_VID = df_vcf.loc[:, ['VID', 'chrom', 'pos', 'ref', 'alt']]

        df_VID.to_csv(self.VID_file, sep='\t', index=False)

    def otsu_threshold(self, array):
        threshold = utils.otsu_min_support_read(array, self.otsu_plot)
        return threshold

    @utils.add_log
    def get_UMI(self):
        '''
        get variant and ref UMI supporting an allele
        '''

        _df_index, df_valid = self.read_CID()

        df_UMI_list = []
        CID_arg = list(df_valid['CID'])
        outdir_arg = [self.outdir] * len(CID_arg)
        df_vcf = parse_vcf(self.merged_vcf_file)
        vcf_arg = [df_vcf] * len(CID_arg)
        with ProcessPoolExecutor(self.thread) as pool:
            for res in pool.map(cell_UMI, CID_arg, outdir_arg, vcf_arg):
                df_UMI_list.append(res)

        df_UMI = pd.concat(df_UMI_list)
        df_UMI['VID'] = df_UMI['VID'].astype('int')
        df_UMI.sort_values(by=['VID', 'CID'], inplace=True)
        df_UMI.to_csv(self.variant_count_file, sep='\t', index=False)

        df_UMI = pd.read_csv(self.variant_count_file, sep='\t', header=0)
        df_filter = df_UMI.copy()
        if self.args.min_support_read == 'auto':
            df_otsu_threshold = pd.DataFrame(columns=['VID', 'threshold'])
            VIDs = np.unique(df_UMI['VID'].values)
            for VID in VIDs:
                df = df_UMI[df_UMI['VID'] == VID]
                df_alt = df[df['alt_count'] > 0]
                array = list(df_alt['alt_count'])
                if array:
                    if len(array) >= OTSU_READ_MIN:
                        min_support_read = utils.otsu_min_support_read(array, f'{self.otsu_dir}/{VID}.png')
                    else:
                        min_support_read = np.median(array) * 0.2
                    df_otsu_threshold = df_otsu_threshold.append(
                        {'VID': VID, 'threshold': min_support_read}, ignore_index=True)
                    df_otsu_threshold.to_csv(self.otsu_threshold_file, sep='\t', index=False)
                    df_filter.loc[((df_filter['VID'] == VID) & (df_filter['alt_count'] < min_support_read)), 'alt_count'] = 0
        else:
            min_support_read = int(self.args.min_support_read)
            df_filter.loc[df_filter['alt_count'] < min_support_read, 'alt_count'] = 0
        df_filter.to_csv(self.filter_variant_count_file, sep='\t', index=False)
   
    def get_bed_df(self):
        bed_df = utils.get_gene_region_from_bed(self.panel)[1]
        return bed_df

    @utils.add_log
    def write_RID_file(self):
        rid_file = self.get_bed_df()
        rid_file.insert(0, 'RID', [rid + 1 for rid in range(len(rid_file))], allow_duplicates=False)
        rid_file.to_csv(self.RID_file,sep = '\t',index = False)
    
    @utils.add_log
    def get_region_dict(self):
        """
        Return a dict
        - key: vid
        - value: rid
        """
        df_rid = pd.read_table(self.RID_file,sep = '\t')
        df_vid = pd.read_table(self.VID_file,sep = '\t')
        
        bed_file_df = self.get_bed_df()
        gr = pr.PyRanges(bed_file_df)
        region_dict = {}
        for _, row in df_vid.iterrows():
            vid = row['VID']
            bed_region = gr[str(row['chrom']), row['pos']: row['pos'] + 1]
            if bed_region:
                bed_chr = bed_region.as_df().loc[:,"Chromosome"].astype(int).to_list()
                bed_start = bed_region.as_df().loc[:,"Start"].to_list()
                bed_end = bed_region.as_df().loc[:,"End"].to_list()
                #get rid
                for chrs, start, end in zip(bed_chr,bed_start,bed_end):
                    rid = df_rid[(df_rid.loc[:,"Chromosome"] == chrs) 
                                 & (df_rid.loc[:,"Start"] == start) 
                                 & (df_rid.loc[:,"End"] == end)].loc[:,"RID"].to_list()[0]
                    region_dict[vid] = rid
            else:
                region_dict[vid] = "None"
        return region_dict
    
    @utils.add_log
    def add_RID(self):
        df_vid = pd.read_table(self.VID_file,sep = '\t')
        df_filter_variant_conut = pd.read_table(self.filter_variant_count_file,sep = '\t')
        
        region_dict = self.get_region_dict()
        df_vid['RID'] = df_vid['VID'].apply(lambda x: region_dict[x])
        df_filter_variant_conut['RID'] = df_filter_variant_conut['VID'].apply(lambda x: region_dict[x])

        df_vid.to_csv(self.VID_file,sep = '\t',index = None)
        df_filter_variant_conut.to_csv(self.filter_variant_count_file,sep = '\t',index = None)
    
    @utils.add_log
    def filter_vcf(self):
        """
        filter cells with zero variant UMI
        """
        df_filter = pd.read_csv(self.filter_variant_count_file, sep='\t')
        df_filter = df_filter[df_filter['alt_count']>0]

        vcf = pysam.VariantFile(self.merged_vcf_file, 'r')
        vcf_header = vcf.header
        filter_vcf = pysam.VariantFile(self.filter_vcf_file, 'w', header=vcf_header)
        for rec in vcf.fetch():
            VID = int(rec.info['VID'])
            CIDs = df_filter[df_filter['VID']==VID]['CID'].values
            rec.info['CID'] = ','.join([str(CID) for CID in CIDs])
            if rec.info['CID']:
                filter_vcf.write(rec)
        filter_vcf.close()
        vcf.close()


    @utils.add_log
    def write_support_matrix(self):
        def set_support_bit(row):
            ref_bit = 1 if row['ref_count'] > 0 else 0
            alt_bit = 2 if row['alt_count'] > 0 else 0
            support_bit = ref_bit + alt_bit
            return support_bit

        df_variant_count = pd.read_csv(self.filter_variant_count_file, sep='\t')
        df_variant_count['support'] = df_variant_count.apply(set_support_bit, axis=1)
        support_mtx = coo_matrix(
            (df_variant_count.support, (df_variant_count.VID - 1, df_variant_count.CID - 1))
        )
        mmwrite(self.support_matrix_file, support_mtx)

    def run(self):
        
        self.SplitNCigarReads()
        self.split_bam()
        self.write_CID_file()
        self.call_all_snp()
        self.merge_vcf()
        self.write_VID_file()
        self.get_UMI()
        if self.panel_bool:
            self.write_RID_file()
            self.add_RID()
        self.filter_vcf()
        self.write_support_matrix()
        self.clean_up()


@utils.add_log
def variant_calling(args):

    step_name = 'variant_calling'
    variant_calling_obj = Variant_calling(args, step_name)
    variant_calling_obj.run()


def get_opts_variant_calling(parser, sub_program):

    parser.add_argument("--genomeDir", help=HELP_DICT['genomeDir'], required=True)
    parser.add_argument(
        "--min_support_read",
        help="""Minimum number of reads support a variant. If `auto`(default), otsu method will be used to determine this value.""",
        default='auto',
    )
    parser.add_argument(
        "--panel",
        help = "The prefix of bed file, such as `lung_1`.",
        default = ''
        )
    if sub_program:
        parser.add_argument(
            "--bam",
            help='Input BAM file from step `target_metrics`. ',
            required=True
        )
        parser.add_argument(
            "--match_dir",
            help=HELP_DICT['match_dir'],
            required=True
        )
        s_common(parser)
